import os
import re
import requests
from typing import Callable
from urllib.parse import quote
from mutagen.id3 import ID3, ID3NoHeaderError, TIT2, TPE1, TALB, TPE2, TDRC, TRCK, TPOS, APIC, TPUB, WXXX, COMM
from mutagen.mp3 import MP3

def sanitize_filename(value: str) -> str:
    return re.sub(r'[\\/*?:"<>|]', "", value).strip()

def safe_int(value) -> int:
    try:
        return int(value)
    except (ValueError, TypeError):
        return 0

class YouTubeDownloader:
    def __init__(self, timeout: float = 120.0):
        self.session = requests.Session()
        self.session.timeout = timeout
        self.session.headers.update({
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/145.0.0.0 Safari/537.36"
        })
        self.progress_callback: Callable[[int, int], None] = None

    def set_progress_callback(self, callback: Callable[[int, int], None]) -> None:
        self.progress_callback = callback

    def get_youtube_url_from_spotify(self, spotify_track_id: str, track_name: str = None, artist_name: str = None) -> str:
        print("Fetching YouTube URL via Songlink HTML...")

        url = f"https://song.link/s/{spotify_track_id}"
        headers = {"User-Agent": "Mozilla/5.0", "Accept-Language": "en-US,en;q=0.9"}

        try:
            # ATTEMPT 1: Try to extract from Songlink
            resp = self.session.get(url, headers=headers, timeout=10)
            resp.raise_for_status()
            html = resp.text

            match = re.search(r'https://(?:music\.)?youtube\.com/watch\?v=([a-zA-Z0-9_-]{11})', html)
            if not match:
                match = re.search(r'https://youtu\.be/([a-zA-Z0-9_-]{11})', html)

            if match:
                video_id = match.group(1)
                yt_url = f"https://music.youtube.com/watch?v={video_id}"
                print(f"✓ Found on Songlink: {yt_url}")
                return yt_url
            else:
                print("[!] Songlink does not have a YouTube link for this track.")

        except Exception as e:
            print(f"[!] Error accessing Songlink: {e}")

        # ATTEMPT 2: Direct YouTube Fallback (Text Search)
        print("Starting direct YouTube search (Fallback)...")
        if track_name and artist_name:
            yt_url = self._search_youtube_direct(track_name, artist_name)
            if yt_url:
                return yt_url
                
        raise Exception("Failed to resolve YouTube URL: Songlink failed and direct search did not find the track.")

    def _search_youtube_direct(self, track_name: str, artist_name: str) -> str:
        """
        Performs a silent search on YouTube and grabs the first video ID.
        This prevents reliance on Songlink for old/famous tracks.
        """
        # "audio" helps filter out music videos with long intros
        query = quote(f"{track_name} {artist_name} audio")
        search_url = f"https://www.youtube.com/results?search_query={query}"
        
        try:
            resp = self.session.get(search_url, timeout=10)
            resp.raise_for_status()
            
            # YouTube stores search data inside a JS variable called ytInitialData
            # This regex captures the first Video ID that appears on the search page
            match = re.search(r'"videoId":"([a-zA-Z0-9_-]{11})"', resp.text)
            if match:
                video_id = match.group(1)
                yt_url = f"https://music.youtube.com/watch?v={video_id}"
                print(f"✓ Video found via YouTube Search: {yt_url}")
                return yt_url
                
        except Exception as e:
            print(f"Error in direct YouTube search: {e}")
            
        return None

    def _extract_video_id(self, url: str) -> str:
        match = re.search(r'(?:v=|/v/|youtu\.be/|/embed/)([a-zA-Z0-9_-]{11})', url)
        return match.group(1) if match else None

    def _request_spotube_dl(self, video_id: str, audio_format="mp3", bitrate="320"):
        engines = ["v1", "v3", "v2"]
        for engine in engines:
            api_url = f"https://spotubedl.com/api/download/{video_id}?engine={engine}&format={audio_format}&quality={bitrate}"
            try:
                print(f"Fetching from SpotubeDL (Engine: {engine})...")
                resp = self.session.get(api_url, timeout=15)
                if resp.status_code == 200:
                    data = resp.json()
                    download_url = data.get("url")
                    
                    if download_url:
                        if download_url.startswith("/"):
                            download_url = "https://spotubedl.com" + download_url
                            
                        return download_url
            except Exception:
                continue
        return None

    def _request_cobalt(self, video_id: str, audio_format="mp3", bitrate="320"):
        print("SpotubeDL failed. Trying Cobalt API (Fallback)...")
        api_url = "https://api.qwkuns.me"
        payload = {
            "url": f"https://music.youtube.com/watch?v={video_id}",
            "audioFormat": audio_format,
            "audioBitrate": str(bitrate),
            "downloadMode": "audio",
            "filenameStyle": "basic",
            "disableMetadata": True
        }
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json"
        }
        try:
            resp = self.session.post(api_url, json=payload, headers=headers, timeout=15)
            if resp.status_code == 200:
                data = resp.json()
                if data.get("status") in ["tunnel", "redirect"] and data.get("url"):
                    return data["url"]
        except Exception:
            pass
        return None

    def download_by_spotify_id(self, spotify_track_id, **kwargs):
        output_dir = kwargs.get("output_dir", ".")
        os.makedirs(output_dir, exist_ok=True)
        
        # Passing names in case Songlink fails!
        yt_url = self.get_youtube_url_from_spotify(
            spotify_track_id,
            track_name=kwargs.get("spotify_track_name"),
            artist_name=kwargs.get("spotify_artist_name")
        )
        
        video_id = self._extract_video_id(yt_url)
        if not video_id:
            raise Exception("Could not extract video ID.")

        safe_title = sanitize_filename(kwargs.get("spotify_track_name", "Unknown"))
        safe_artist = sanitize_filename(kwargs.get("spotify_artist_name", "Unknown").split(",")[0])
        
        expected_filename = f"{safe_artist} - {safe_title}.mp3" 
        expected_path = os.path.join(output_dir, expected_filename)

        if os.path.exists(expected_path) and os.path.getsize(expected_path) > 0:
            print(f"File already exists: {expected_path}")
            return expected_path

        download_url = self._request_spotube_dl(video_id, "mp3", "320")
        if not download_url:
            download_url = self._request_cobalt(video_id, "mp3", "320")
            
        if not download_url:
            raise Exception("All YouTube download APIs failed.")

        print("Downloading track from YouTube...")
        with self.session.get(download_url, stream=True) as r:
            r.raise_for_status()
            total = int(r.headers.get("Content-Length") or 0)
            downloaded = 0
            with open(expected_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
                        downloaded += len(chunk)
                        if self.progress_callback:
                            self.progress_callback(downloaded, total)
        print()

        self.embed_metadata(
            expected_path, 
            kwargs.get("spotify_track_name"), kwargs.get("spotify_artist_name"),
            kwargs.get("spotify_album_name"), kwargs.get("spotify_album_artist"),
            kwargs.get("spotify_release_date"), kwargs.get("spotify_track_number", 1),
            kwargs.get("spotify_total_tracks", 1), kwargs.get("spotify_disc_number", 1),
            kwargs.get("spotify_total_discs", 1), kwargs.get("spotify_cover_url"),
            kwargs.get("spotify_publisher"), kwargs.get("spotify_url") 
        )

        return expected_path

    def embed_metadata(self, filepath, title, artist, album, album_artist, date, track_num, total_tracks, disc_num, total_discs, cover_url, publisher=None, url=None):
        print("Embedding metadata and cover art...")
        try:
            try:
                audio = ID3(filepath)
                audio.delete() 
            except ID3NoHeaderError:
                audio = ID3()
            
            if title: audio.add(TIT2(encoding=3, text=str(title)))
            if artist: audio.add(TPE1(encoding=3, text=str(artist)))
            if album: audio.add(TALB(encoding=3, text=str(album)))
            if album_artist: audio.add(TPE2(encoding=3, text=str(album_artist)))
            if date: audio.add(TDRC(encoding=3, text=str(date)))
            
            audio.add(TRCK(encoding=3, text=f"{safe_int(track_num)}/{safe_int(total_tracks)}"))
            audio.add(TPOS(encoding=3, text=f"{safe_int(disc_num)}/{safe_int(total_discs)}"))

            if publisher: 
                audio.add(TPUB(encoding=3, text=[str(publisher)]))
            if url: 
                audio.add(WXXX(encoding=3, desc=u'', url=str(url)))
                
            audio.add(COMM(
                encoding=3, 
                lang='eng', 
                desc=u'', 
                text=[u"https://github.com/ShuShuzinhuu/SpotiFLAC-Module-Version"]
            ))

            if cover_url:
                try:
                    resp = self.session.get(cover_url, timeout=10)
                    if resp.status_code == 200:
                        audio.add(APIC(
                            encoding=3, 
                            mime='image/jpeg', 
                            type=3, 
                            desc='Cover', 
                            data=resp.content
                        ))
                except Exception as e:
                    print(f"Warning: Could not download cover: {e}")

            audio.save(filepath, v2_version=3)
            print("Metadata embedded successfully")
            
        except Exception as e:
            print(f"Warning: Failed to embed metadata: {e}")


import json
import os
import re
import random
import time
from typing import Callable, Dict

import requests
from mutagen.flac import FLAC, Picture
from mutagen.id3 import PictureType

def _sanitize_filename(value: str, fallback: str = "Unknown") -> str:
    if not value:
        return fallback
    cleaned = re.sub(r'[\\/*?:"<>|]', "", value)
    cleaned = re.sub(r"\s+", " ", cleaned).strip()
    return cleaned or fallback

def get_first_artist(artist_str: str) -> str:
    if not artist_str:
        return "Unknown"
    return artist_str.split(",")[0].strip()

def build_qobuz_filename(title, artist, album, album_artist, release_date, track_number, disc_number, format_string, include_track_number, position, use_album_track_number):
    number_to_use = track_number if use_album_track_number and track_number > 0 else position
    year = release_date[:4] if len(release_date) >= 4 else ""

    if "{" in format_string:
        filename = (format_string.replace("{title}", title)
                    .replace("{artist}", artist)
                    .replace("{album}", album)
                    .replace("{album_artist}", album_artist)
                    .replace("{year}", year)
                    .replace("{date}", _sanitize_filename(release_date)))
        
        if disc_number > 0:
            filename = filename.replace("{disc}", str(disc_number))
        else:
            filename = filename.replace("{disc}", "")
            
        if number_to_use > 0:
            filename = filename.replace("{track}", f"{number_to_use:02d}")
        else:
            filename = re.sub(r"\{track\}[\.\s-]*", "", filename)
    else:
        if format_string == "artist-title":
            filename = f"{artist} - {title}"
        elif format_string == "title":
            filename = title
        else:
            filename = f"{title} - {artist}"
        if include_track_number and position > 0:
            filename = f"{number_to_use:02d}. {filename}"

    return _sanitize_filename(filename) + ".flac"

def build_qobuz_api_url(api_base: str, track_id: int, quality: str) -> str:
    if "qbz.afkarxyz.fun" in api_base:
        return f"{api_base}{track_id}?quality={quality}"
    return f"{api_base}{track_id}&quality={quality}"

class QobuzDownloader:
    def __init__(self, timeout: float = 60.0, app_id: str = "798273057"):
        self.timeout = timeout
        self.app_id = app_id
        self.session = requests.Session()
        self.session.timeout = timeout
        self.progress_callback = lambda current, total: None

    def set_progress_callback(self, callback: Callable[[int, int], None]) -> None:
        self.progress_callback = callback

    def _search_by_isrc(self, isrc: str) -> Dict:
        api_base = "https://www.qobuz.com/api.json/0.2/track/search?query="
        url = f"{api_base}{isrc}&limit=1&app_id={self.app_id}"
        
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/145.0.0.0 Safari/537.36"
        }
        
        resp = self.session.get(url, headers=headers)
        if resp.status_code != 200:
             try:
                 err_msg = resp.json().get("message", f"Status {resp.status_code}")
             except:
                 err_msg = f"Status {resp.status_code}"
             raise Exception(f"API Error: {err_msg}")

        data = resp.json()
        items = data.get("tracks", {}).get("items", [])
        if not items:
            raise Exception(f"track not found for ISRC: {isrc}")
        return items[0]

    def _download_from_standard(self, api_base: str, track_id: int, quality: str) -> str:
        url = build_qobuz_api_url(api_base, track_id, quality)
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36"
        }
        resp = self.session.get(url, headers=headers, timeout=self.timeout)
        
        if resp.status_code != 200:
            raise Exception(f"status {resp.status_code}")
        
        if not resp.text.strip():
             raise Exception("empty body")

        try:
            data = resp.json()
        except Exception:
            raise Exception("invalid response")

        if isinstance(data, dict):
            if data.get("url"):
                return data["url"]
            if data.get("data", {}).get("url"):
                return data["data"]["url"]
        
        raise Exception("invalid response")

    def get_download_url(self, track_id: int, quality: str, allow_fallback: bool) -> str:
        quality_code = quality if quality not in ["", "5"] else "6"
        
        print(f"Getting download URL for track ID: {track_id} with requested quality: {quality_code}")

        standard_apis = [
            "https://dab.yeet.su/api/stream?trackId=",
            "https://dabmusic.xyz/api/stream?trackId=",
            "https://qbz.afkarxyz.fun/api/track/"
        ]

        def attempt_download(qual):
            providers = []
            for api in standard_apis:
                providers.append({"name": f"Standard({api})", "func": lambda a=api: self._download_from_standard(a, track_id, qual)})
            
            random.seed(time.time())
            random.shuffle(providers)
            
            last_err = None
            for p in providers:
                print(f"Trying Provider: {p['name']} (Quality: {qual})...")
                try:
                    url = p['func']()
                    if url:
                        print("✓ Success")
                        return url
                except Exception as e:
                    print(f"Provider failed: {e}")
                    last_err = e
                    
            raise Exception(last_err)

        try:
            url = attempt_download(quality_code)
            if url: return url
        except Exception as e:
            err = e

        if allow_fallback:
            current_qual = quality_code
            
            if current_qual == "27":
                print(f"⚠ Download with quality 27 failed, trying fallback to 7 (24-bit Standard)...")
                try:
                    url = attempt_download("7")
                    if url: 
                        print("✓ Success with fallback quality 7")
                        return url
                except:
                    pass
                current_qual = "7"

            if current_qual == "7":
                print(f"⚠ Download with quality 7 failed, trying fallback to 6 (16-bit Lossless)...")
                try:
                    url = attempt_download("6")
                    if url: 
                        print("✓ Success with fallback quality 6")
                        return url
                except:
                    pass
        
        raise Exception("all APIs and fallbacks failed")

    def _stream_download(self, url: str, filepath: str) -> None:
        print("Starting file download...")
        os.makedirs(os.path.dirname(filepath) or ".", exist_ok=True)
        temp_path = filepath + ".part"
        
        try:
            with self.session.get(url, stream=True, timeout=300) as resp:
                if resp.status_code != 200:
                    raise Exception(f"download failed with status {resp.status_code}")
                
                print(f"Creating file: {filepath}")
                print("Downloading...")
                
                total = int(resp.headers.get("Content-Length") or 0)
                downloaded = 0
                with open(temp_path, "wb") as f:
                    for chunk in resp.iter_content(chunk_size=256 * 1024):
                        if chunk:
                            f.write(chunk)
                            downloaded += len(chunk)
                            if self.progress_callback:
                                self.progress_callback(downloaded, total)
                                
            os.replace(temp_path, filepath)
            print(f"\rDownloaded: {downloaded / (1024 * 1024):.2f} MB (Complete)")
        finally:
            if os.path.exists(temp_path):
                try: os.remove(temp_path)
                except: pass

    def download_by_isrc(self, isrc, output_dir, quality, filename_format, include_track_number, position, 
                         spotify_track_name, spotify_artist_name, spotify_album_name, use_album_track_number,
                         **kwargs):
        
        spotify_album_artist = kwargs.get("spotify_album_artist", "Unknown")
        spotify_release_date = kwargs.get("spotify_release_date", "")
        spotify_track_number = kwargs.get("spotify_track_number", 0)
        spotify_disc_number = kwargs.get("spotify_disc_number", 1)
        spotify_total_tracks = kwargs.get("spotify_total_tracks", 0)
        spotify_total_discs = kwargs.get("spotify_total_discs", 1)
        spotify_cover_url = kwargs.get("spotify_cover_url", "")
        spotify_copyright = kwargs.get("spotify_copyright", "")
        spotify_publisher = kwargs.get("spotify_publisher", "")
        spotify_url = kwargs.get("spotify_url", "")
        allow_fallback = kwargs.get("allow_fallback", True)
        use_first_artist_only = kwargs.get("use_first_artist_only", False)

        print(f"Fetching track info for ISRC: {isrc}")
        if output_dir != ".":
            os.makedirs(output_dir, exist_ok=True)

        track = self._search_by_isrc(isrc)
        
        q_track_num = track.get("track_number", 0)
        final_track_num = q_track_num if (use_album_track_number and q_track_num > 0) else position
        if final_track_num == 0 and spotify_track_number > 0:
            final_track_num = spotify_track_number

        print(f"Found track: {spotify_artist_name} - {spotify_track_name}")
        print(f"Album: {spotify_album_name}")
        
        if track.get('hires', False):
            print(f"Quality: Hi-Res ({track.get('maximum_bit_depth', 24)}-bit / {track.get('maximum_sampling_rate', 96.0)} kHz)")
        else:
            print("Quality: Standard")

        # Artist optimization match
        artist_to_use = get_first_artist(spotify_artist_name) if use_first_artist_only else spotify_artist_name
        album_artist_to_use = get_first_artist(spotify_album_artist) if use_first_artist_only else spotify_album_artist
        
        filename = build_qobuz_filename(
            _sanitize_filename(spotify_track_name), 
            _sanitize_filename(artist_to_use), 
            _sanitize_filename(spotify_album_name), 
            _sanitize_filename(album_artist_to_use),
            spotify_release_date, final_track_num, spotify_disc_number, 
            filename_format, include_track_number, position, use_album_track_number
        )
        filepath = os.path.join(output_dir, filename)

        if os.path.exists(filepath) and os.path.getsize(filepath) > 0:
            size_mb = os.path.getsize(filepath) / (1024 * 1024)
            print(f"File already exists: {filepath} ({size_mb:.2f} MB)")
            return filepath

        print("Getting download URL...")
        download_url = self.get_download_url(track['id'], quality, allow_fallback)
        
        url_preview = download_url[:60] + "..." if len(download_url) > 60 else download_url
        print(f"Download URL obtained: {url_preview}")
        
        print(f"Downloading FLAC file to: {filepath}")
        self._stream_download(download_url, filepath)
        print(f"Downloaded: {filepath}")

        cover_path = ""
        if spotify_cover_url:
            cover_path = filepath + ".cover.jpg"
            try:
                with open(cover_path, "wb") as f:
                    f.write(self.session.get(spotify_cover_url, timeout=15).content)
                print("Spotify cover downloaded")
            except Exception as e:
                print(f"Warning: Failed to download Spotify cover: {e}")
                cover_path = ""

        print("Embedding metadata and cover art...")
        metadata = {
            "TITLE": spotify_track_name,
            "ARTIST": spotify_artist_name,
            "ALBUM": spotify_album_name,
            "ALBUMARTIST": spotify_album_artist,
            "DATE": spotify_release_date[:4] if len(spotify_release_date) >= 4 else "",
            "TRACKNUMBER": str(final_track_num),
            "TRACKTOTAL": str(spotify_total_tracks),
            "DISCNUMBER": str(spotify_disc_number),
            "DISCTOTAL": str(spotify_total_discs),
            "ISRC": isrc,
            "COPYRIGHT": spotify_copyright,
            "ORGANIZATION": spotify_publisher,
            "URL": spotify_url,
            "DESCRIPTION": "https://github.com/ShuShuzinhuu/SpotiFLAC-Module-Version"
        }

        self._embed_metadata(filepath, metadata, cover_path)
        
        if cover_path and os.path.exists(cover_path):
            try: os.remove(cover_path)
            except: pass

        return filepath

    def _embed_metadata(self, filepath, metadata, cover_path):
        try:
            audio = FLAC(filepath)
            audio.delete() # Limpa as tags originais que a Qobuz envia
            
            for key, val in metadata.items():
                if val and str(val) != "0": 
                    audio[key] = str(val)
            
            if cover_path and os.path.exists(cover_path):
                with open(cover_path, "rb") as img:
                    pic = Picture()
                    pic.data = img.read()
                    pic.type = PictureType.COVER_FRONT
                    pic.mime = "image/jpeg"
                    audio.add_picture(pic)
            
            audio.save()
            print("Metadata embedded successfully!")
        except Exception as e:
            raise Exception(f"failed to embed metadata: {e}")


import os
import re
from tempfile import template
import time
import argparse
import asyncio
import shutil 
from dataclasses import dataclass, field
from SpotiFLAC.youtubeDL import YouTubeDownloader
from SpotiFLAC.getMetadata import get_filtered_data, parse_uri, SpotifyInvalidUrlException
from SpotiFLAC.tidalDL import TidalDownloader
from SpotiFLAC.deezerDL import DeezerDownloader
from SpotiFLAC.qobuzDL import QobuzDownloader
from SpotiFLAC.amazonDL import AmazonDownloader
from SpotiFLAC.progress import DownloadManager, RichProgressCallback
from SpotiFLAC.check_update import check_for_updates
from SpotiFLAC.spotidownloaderDL import SpotiDownloader


@dataclass
class Config:
    url: str
    output_dir: str
    service: list = None
    filename_format: str = "{title} - {artist}"
    use_track_numbers: bool = False
    use_artist_subfolders: bool = False
    use_album_subfolders: bool = False
    is_album: bool = False
    is_playlist: bool = False
    is_single_track: bool = False
    album_or_playlist_name: str = ""
    tracks: list = field(default_factory=list)
    worker: object = None
    loop: int = 3600
    start_time: float = 0.0
    end_time: float = 0.0


@dataclass
class Track:
    external_urls: str
    title: str
    artists: str
    album: str
    album_artist: str 
    track_number: int
    duration_ms: int
    id: str
    isrc: str = ""
    release_date: str = ""
    cover_url: str = ""
    downloaded: bool = False

def extract_cover_art(data, key_primary="images", key_secondary="album"):
    img_data = data.get(key_primary)
    
    if img_data and isinstance(img_data, str):
        return img_data
        
    if img_data and isinstance(img_data, list) and len(img_data) > 0:
        if isinstance(img_data[0], dict):
            return img_data[0].get("url", "")
        if isinstance(img_data[0], str):
            return img_data[0]

    if key_secondary and key_secondary in data:
        album_data = data[key_secondary]
        if isinstance(album_data, dict):
            return extract_cover_art(album_data, "images", None)
            
    return ""


def format_artists(artists_list):
    if isinstance(artists_list, list):
        return ", ".join([a.get("name", "Unknown") if isinstance(a, dict) else str(a) for a in artists_list])
    return str(artists_list) if artists_list else "Unknown Artist"


def get_metadata(url):
    try:
        metadata = get_filtered_data(url)
        if "error" in metadata:
            print("Error fetching metadata:", metadata["error"])
        else:
            print("Metadata fetched successfully.")
            return metadata
    except SpotifyInvalidUrlException as e:
        print("Invalid URL:", str(e))
    except Exception as e:
        print("An error occurred while fetching metadata:", str(e))


def fetch_tracks(url):
    if not url:
        print('Warning: Please enter a Spotify URL.')
        return

    try:
        print('Just a moment. Fetching metadata...')
        metadata = get_metadata(url)
        if metadata:
            on_metadata_fetched(metadata)
        else:
            print("Error: Empty metadata received.")
    except Exception as e:
        print(f'Error: Failed to start metadata fetch: {str(e)}')


def on_metadata_fetched(metadata):
    try:
        url_info = parse_uri(config.url)

        if url_info["type"] == "track":
            data = metadata.get("track", metadata)
            handle_track_metadata(data)
        elif url_info["type"] == "album":
            handle_album_metadata(metadata)
        elif url_info["type"] == "playlist":
            handle_playlist_metadata(metadata)

    except Exception as e:
        import traceback
        traceback.print_exc()
        print(f'Error parsing metadata: {str(e)}')


def handle_track_metadata(track_data):
    track_id = track_data.get("id")
    if not track_id and "external_urls" in track_data:
        ext = track_data["external_urls"]
        if isinstance(ext, dict): track_id = ext.get("spotify", "").split("/")[-1]
        elif isinstance(ext, str): track_id = ext.split("/")[-1]

    if not track_id:
        print("[!] Skipping track without ID")
        return

    cover = extract_cover_art(track_data)
    
    artist_names = format_artists(track_data.get("artists", []))
    
    album_obj = track_data.get("album", {})
    if isinstance(album_obj, dict) and album_obj.get("artists"):
        album_artist = format_artists(album_obj.get("artists"))
    else:
        album_artist = artist_names

    track = Track(
        external_urls=f"https://open.spotify.com/track/{track_id}",
        title=track_data.get("name", "Unknown Title"),
        artists=artist_names,
        album=track_data.get("album_name", track_data.get("album", {}).get("name", "Unknown Album")),
        album_artist=album_artist,
        track_number=track_data.get("track_number", 1),
        duration_ms=track_data.get("duration_ms", 0),
        id=track_id,
        isrc=track_data.get("external_ids", {}).get("isrc", "") or track_data.get("isrc", ""),
        release_date=track_data.get("album", {}).get("release_date", "") or track_data.get("release_date", ""),
        cover_url=cover 
    )

    config.tracks = [track]
    config.is_single_track = True
    config.is_album = config.is_playlist = False
    config.album_or_playlist_name = f"{config.tracks[0].title} - {config.tracks[0].artists}"


def handle_album_metadata(album_data):
    config.album_or_playlist_name = album_data.get("album_info", {}).get("name", album_data.get("name", "Unknown Album"))
    album_release_date = album_data.get("album_info", {}).get("release_date", album_data.get("release_date", ""))
    
    raw_album_artists = album_data.get("album_info", {}).get("artists", [])
    if not raw_album_artists:
         raw_album_artists = album_data.get("artists", [])
    
    if isinstance(raw_album_artists, str):
        main_album_artist = raw_album_artists
    else:
        main_album_artist = format_artists(raw_album_artists)

    album_cover = extract_cover_art(album_data.get("album_info", album_data))
    tracks_raw = album_data.get("track_list", album_data.get("tracks", {}).get("items", []))

    for track in tracks_raw:
        track_id = track.get("id")
        if not track_id and "external_urls" in track:
            ext = track["external_urls"]
            if isinstance(ext, dict): track_id = ext.get("spotify", "").split("/")[-1]
            elif isinstance(ext, str): track_id = ext.split("/")[-1]

        if not track_id or any(t.id == track_id for t in config.tracks):
            continue

        track_cover = extract_cover_art(track)
        if not track_cover:
            track_cover = album_cover

        artist_names = format_artists(track.get("artists", []))

        config.tracks.append(Track(
            external_urls=f"https://open.spotify.com/track/{track_id}",
            title=track.get("name", "Unknown Title"),
            artists=artist_names,
            album=config.album_or_playlist_name,
            album_artist=main_album_artist,
            track_number=track.get("track_number", 0),
            duration_ms=track.get("duration_ms", 0),
            id=track_id,
            isrc=track.get("isrc", ""), 
            release_date=album_release_date,
            cover_url=track_cover
        ))

    config.is_album = True
    config.is_playlist = config.is_single_track = False


def handle_playlist_metadata(playlist_data):
    info = playlist_data.get("playlist_info", playlist_data)
    config.album_or_playlist_name = info.get("name", "Unknown Playlist")
    
    playlist_cover = extract_cover_art(info)
    
    tracks_raw = playlist_data.get("track_list", [])
    if not tracks_raw and "tracks" in playlist_data:
        tracks_raw = playlist_data["tracks"].get("items", [])

    for item in tracks_raw:
        track = item.get("track", item)
        if not track: continue 
        
        track_id = track.get("id")
        if not track_id and "external_urls" in track:
            ext = track["external_urls"]
            if isinstance(ext, dict): track_id = ext.get("spotify", "").split("/")[-1]
            elif isinstance(ext, str): track_id = ext.split("/")[-1]
            
        if not track_id or any(t.id == track_id for t in config.tracks):
            continue
        
        track_cover = extract_cover_art(track)
        if not track_cover:
            track_cover = playlist_cover
        
        artist_names = format_artists(track.get("artists", []))
        
        alb = track.get("album", {})
        album_name = alb.get("name", track.get("album_name", "Unknown Album"))
        
        if alb.get("artists"):
            album_artist = format_artists(alb.get("artists"))
        else:
            album_artist = artist_names

        release_date = alb.get("release_date", "")

        config.tracks.append(Track(
            external_urls=f"https://open.spotify.com/track/{track_id}",
            title=track.get("name", "Unknown Title"),
            artists=artist_names,
            album=album_name,
            album_artist=album_artist,
            track_number=track.get("track_number", len(config.tracks) + 1),
            duration_ms=track.get("duration_ms", 0),
            id=track_id,
            isrc=track.get("isrc", ""),
            release_date=release_date,
            cover_url=track_cover
        ))

    config.is_playlist = True
    config.is_album = config.is_single_track = False


def download_tracks(indices):
    if not config.tracks:
        print("No tracks found to download.")
        return

    raw_outpath = config.output_dir
    outpath = os.path.normpath(raw_outpath)
    if not os.path.exists(outpath):
        print('Warning: Invalid output directory. Please check if the folder exists.')
        return

    tracks_to_download = config.tracks if config.is_single_track else [config.tracks[i] for i in indices]

    if config.is_album or config.is_playlist:
        name = config.album_or_playlist_name.strip()
        folder_name = re.sub(r'[<>:"/\\|?*]', '_', name)
        outpath = os.path.join(outpath, folder_name)
        os.makedirs(outpath, exist_ok=True)

    manager = DownloadManager()
    for track in tracks_to_download:
        manager.add_to_queue(
            item_id=track.id,
            track_name=track.title,
            artist_name=track.artists,
            album_name=track.album,
            spotify_id=track.id
        )

    try:
        start_download_worker(tracks_to_download, outpath)
    except Exception as e:
        import traceback
        traceback.print_exc()
        print(f"Error starting download: {str(e)}")


def start_download_worker(tracks_to_download, outpath):
    config.worker = DownloadWorker(
        tracks_to_download,
        outpath,
        config.is_single_track,
        config.is_album,
        config.is_playlist,
        config.album_or_playlist_name,
        config.filename_format,
        config.use_track_numbers,
        config.use_artist_subfolders,
        config.use_album_subfolders,
        config.service,
    )
    config.worker.run()


def on_download_finished(success, message, failed_tracks, total_elapsed=None):
    if success:
        print(f"\n=======================================")
        print(f"\nStatus: {message}")
        if failed_tracks:
            print("\nFailed downloads:")
            for title, artists, error in failed_tracks:
                print(f"• {title} - {artists}")
                print(f"  Error: {error}\n")
    else:
        print(f"Error: {message}")

    if total_elapsed is not None:
        print(f"\nElapsed time for this download loop: {format_seconds(total_elapsed)}")

    if config.loop is not None and config.loop > 0:
        print(f"\nDownload starting again in: {format_minutes(config.loop)}")
        print(f"\n=======================================")
        time.sleep(config.loop * 60)
        fetch_tracks(config.url)
        download_tracks(range(len(config.tracks)))


def update_progress(message):
    print(message)


def format_minutes(minutes):
    if not isinstance(minutes, (int, float)):
        return f"{minutes} (invalid format)"
        
    if minutes < 60:
        return f"{minutes} minutes"
    elif minutes < 1440:
        hours = minutes // 60
        mins = minutes % 60
        return f"{hours} hours {mins} minutes"
    else:
        days = minutes // 1440
        hours = (minutes % 1440) // 60
        mins = minutes % 60
        return f"{days} days {hours} hours {mins} minutes"


def format_seconds(seconds: float) -> str:
    seconds = int(round(seconds))
    days, rem = divmod(seconds, 86400)
    hrs, rem = divmod(rem, 3600)
    mins, secs = divmod(rem, 60)
    parts = []
    if days: parts.append(f"{days}d")
    if hrs: parts.append(f"{hrs}h")
    if mins: parts.append(f"{mins}m")
    if secs or not parts: parts.append(f"{secs}s")
    return " ".join(parts)


def sanitize_filename_component(value: str) -> str:
    if not value: return ""
    sanitized = re.sub(r'[<>:"/\\|?*]', lambda m: "'" if m.group() == '"' else '_', value)
    sanitized = re.sub(r'\s+', ' ', sanitized).strip()
    return sanitized


def format_custom_filename(template: str, track, position: int = 1, ext: str = ".flac") -> str:
    year = ""
    if track.release_date:
        year = track.release_date.split("-")[0] if "-" in track.release_date else track.release_date

    duration = ""
    if track.duration_ms:
        total_seconds = track.duration_ms // 1000
        minutes = total_seconds // 60
        seconds = total_seconds % 60
        duration = f"{minutes:02d}:{seconds:02d}"

    replacements = {
        "title": sanitize_filename_component(track.title),
        "artist": sanitize_filename_component(track.artists),
        "album": sanitize_filename_component(track.album),
        "track_number": f"{track.track_number:02d}" if track.track_number else f"{position:02d}",
        "track": f"{track.track_number:02d}" if track.track_number else f"{position:02d}",
        "date": sanitize_filename_component(track.release_date),
        "year": year,
        "position": f"{position:02d}",
        "isrc": sanitize_filename_component(track.isrc),
        "duration": duration,
    }

    result = template
    for key, value in replacements.items():
        result = result.replace(f"{{{key}}}", value)

    # Troque a checagem fixa pela extensão que veio no parâmetro
    if not result.lower().endswith(ext):
        result += ext
    return re.sub(r'\s+', ' ', result).strip()


class DownloadWorker:
    def __init__(self, tracks, outpath, is_single_track=False, is_album=False, is_playlist=False,
                 album_or_playlist_name='', filename_format='{title} - {artist}', use_track_numbers=True,
                 use_artist_subfolders=False, use_album_subfolders=False, services=["tidal"]):
        super().__init__()
        self.tracks = tracks
        self.outpath = outpath
        self.is_single_track = is_single_track
        self.is_album = is_album
        self.is_playlist = is_playlist
        self.album_or_playlist_name = album_or_playlist_name
        self.filename_format = filename_format
        self.use_track_numbers = use_track_numbers
        self.use_artist_subfolders = use_artist_subfolders
        self.use_album_subfolders = use_album_subfolders
        self.services = services
        self.failed_tracks = []

    def get_formatted_filename(self, track, position=1, ext=".flac"):
        if self.filename_format in ["title_artist", "artist_title", "title_only"]:
            if self.filename_format == "artist_title":
                filename = f"{track.artists} - {track.title}{ext}"
            elif self.filename_format == "title_only":
                filename = f"{track.title}{ext}"
            else:
                filename = f"{track.title} - {track.artists}{ext}"
            return re.sub(r'[<>:"/\\|?*]', lambda m: "'" if m.group() == '"' else '_', filename)
        return format_custom_filename(self.filename_format, track, position, ext)

    def run(self):
        try:
            total_tracks = len(self.tracks)
            start = time.perf_counter()
            manager = DownloadManager() 

            for i, track in enumerate(self.tracks):
                if track.downloaded: 
                    manager.complete_download(track.id, "Already downloaded", 0.0)
                    continue

                update_progress(f"\n[{i + 1}/{total_tracks}] Starting download: {track.title} - {track.artists}")
                
                manager.start_download(track.id)

                track_outpath = self.outpath
                if self.is_playlist:
                    if self.use_artist_subfolders:
                        artist_folder = re.sub(r'[<>:"/\\|?*]', '_', track.artists.split(", ")[0])
                        track_outpath = os.path.join(track_outpath, artist_folder)
                    if self.use_album_subfolders:
                        album_folder = re.sub(r'[<>:"/\\|?*]', '_', track.album)
                        track_outpath = os.path.join(track_outpath, album_folder)
                    os.makedirs(track_outpath, exist_ok=True)

                new_filename = self.get_formatted_filename(track, i + 1)
                new_filepath = os.path.join(track_outpath, new_filename)

                if os.path.exists(new_filepath) and os.path.getsize(new_filepath) > 0:
                    update_progress(f"File already exists: {new_filename}. Skipping download.")
                    track.downloaded = True
                    size_mb = os.path.getsize(new_filepath) / (1024 * 1024)
                    manager.complete_download(track.id, new_filepath, size_mb)
                    continue

                download_success = False
                last_error = None

                for svc in self.services:
                    update_progress(f"Trying service: {svc}")
                    
                    current_ext = ".mp3" if svc == "youtube" else ".m4a" if svc == "amazon" else ".flac"
                    
                    new_filename = self.get_formatted_filename(track, i + 1, current_ext)
                    new_filepath = os.path.join(track_outpath, new_filename)
                    
                    if svc == "tidal": downloader = TidalDownloader()
                    elif svc == "deezer": downloader = DeezerDownloader()
                    elif svc == "qobuz": downloader = QobuzDownloader()
                    elif svc == "amazon": downloader = AmazonDownloader()
                    elif svc == "youtube": downloader = YouTubeDownloader()
                    elif svc == "spoti": downloader = SpotiDownloader()
                    else: downloader = TidalDownloader()

                    progress_cb = RichProgressCallback(item_id=track.id)
                    if hasattr(downloader, "set_progress_callback"):
                        downloader.set_progress_callback(progress_cb)

                    try:
                        downloaded_file = None
                        
                        if svc == "tidal":
                            downloaded_file = downloader.download_by_spotify_id(
                                spotify_track_id=track.id,
                                isrc=track.isrc,
                                output_dir=track_outpath,
                                filename_format=self.filename_format,
                                include_track_number=self.use_track_numbers,
                                position=track.track_number or i + 1,
                                spotify_track_name=track.title,
                                spotify_artist_name=track.artists,
                                spotify_album_name=track.album,
                                spotify_album_artist=track.album_artist,
                                spotify_release_date=track.release_date,
                                use_album_track_number=self.use_track_numbers,
                                spotify_cover_url=track.cover_url
                            )
                        elif svc == "deezer":
                            if not track.isrc: raise Exception("No ISRC for Deezer")
                            ok = asyncio.run(downloader.download_by_isrc(track.isrc, track_outpath))
                            if not ok: raise Exception("Deezer download failed")
                            import glob
                            flac_files = glob.glob(os.path.join(track_outpath, "*.flac"))
                            if flac_files: downloaded_file = max(flac_files, key=os.path.getctime)
                        elif svc == "qobuz":
                            if not track.isrc: raise Exception("No ISRC for Qobuz")
                            downloaded_file = downloader.download_by_isrc(
                                isrc=track.isrc,
                                output_dir=track_outpath,
                                quality="6",
                                filename_format=self.filename_format.replace("{title}", "temp_qobuz").replace("{artist}", "temp"),
                                include_track_number=False,
                                position=track.track_number or i + 1,
                                spotify_track_name=track.title,
                                spotify_artist_name=track.artists,
                                spotify_album_name=track.album,
                                spotify_album_artist=track.album_artist,
                                spotify_release_date=track.release_date, 
                                use_album_track_number=self.use_track_numbers,
                                spotify_cover_url=track.cover_url
                            )
                        elif svc == "amazon":
                            downloaded_file = downloader.download_by_spotify_id(
                                spotify_track_id=track.id,
                                output_dir=track_outpath,
                                isrc=track.isrc,
                                filename_format="temp_amazon",
                                include_track_number=self.use_track_numbers,
                                position=track.track_number or i + 1,
                                spotify_track_name=track.title,
                                spotify_artist_name=track.artists,
                                spotify_album_name=track.album,
                                spotify_album_artist=track.album_artist, 
                                spotify_release_date=track.release_date, 
                                use_album_track_number=self.use_track_numbers,
                                spotify_cover_url=track.cover_url
                            )
                        elif svc == "youtube":
                            downloaded_file = downloader.download_by_spotify_id(
                                spotify_track_id=track.id,
                                output_dir=track_outpath,
                                spotify_track_name=track.title,
                                spotify_artist_name=track.artists,
                                spotify_album_name=track.album,
                                spotify_album_artist=track.album_artist, 
                                spotify_release_date=track.release_date, 
                                spotify_track_number=track.track_number or i + 1,
                                spotify_total_tracks=len(self.tracks),
                                spotify_disc_number=1,
                                spotify_total_discs=1,
                                spotify_cover_url=track.cover_url
                            )
                        elif svc == "spoti":
                            downloaded_file = downloader.download_by_spotify_id(
                                spotify_track_id=track.id,
                                output_dir=track_outpath,
                                spotify_track_name=track.title,
                                spotify_artist_name=track.artists,
                                spotify_album_name=track.album,
                                spotify_album_artist=track.album_artist,
                                spotify_release_date=track.release_date,
                                spotify_track_number=track.track_number or i + 1,
                                spotify_total_tracks=len(self.tracks),
                                spotify_disc_number=1,
                                spotify_total_discs=1,
                                spotify_cover_url=track.cover_url
                            )

                        if downloaded_file and os.path.exists(downloaded_file):
                            if downloaded_file != new_filepath:
                                try:
                                    if os.path.exists(new_filepath): os.remove(new_filepath)
                                    shutil.move(downloaded_file, new_filepath)
                                except OSError as e:
                                    print() 
                                    update_progress(f"[!] Rename/Move failed: {e}")
                            
                            print()
                            update_progress(f"Successfully downloaded using: {svc}")
                            track.downloaded = True
                            download_success = True
                            
                            final_size = os.path.getsize(new_filepath) / (1024 * 1024)
                            manager.complete_download(track.id, new_filepath, final_size)
                            
                            break
                        else:
                            raise Exception("File missing after download")

                    except Exception as e:
                        last_error = str(e)
                        print()
                        update_progress(f"[X] {svc} failed: {e}")
                        continue

                if not download_success:
                    self.failed_tracks.append((track.title, track.artists, last_error))
                    print()
                    update_progress(f"[X] Failed all services")
                    manager.fail_download(track.id, last_error or "All services failed")

                if i < total_tracks - 1:
                    time.sleep(1.5)

            total_elapsed = time.perf_counter() - start
            on_download_finished(True, "Download completed!", self.failed_tracks, total_elapsed)

        except Exception as e:
            on_download_finished(False, str(e), self.failed_tracks)


def parse_args():
    parser = argparse.ArgumentParser(
        prog="spotiflac",
        usage="spotiflac url output_dir [-h] [--service {tidal,deezer,qobuz,amazon} ...] "
              "[--filename-format FILENAME_FORMAT] [--use-track-numbers] "
              "[--use-artist-subfolders] [--use-album-subfolders] [--loop LOOP]"
    )

    parser.add_argument("url", help="Spotify URL")
    parser.add_argument("output_dir", help="Output directory")
    parser.add_argument("--service", choices=["tidal", "deezer", "qobuz", "amazon", "youtube", "spoti"], nargs="+", default=["tidal"])
    parser.add_argument("--filename-format", default="{title} - {artist}")
    parser.add_argument("--use-track-numbers", action="store_true")
    parser.add_argument("--use-artist-subfolders", action="store_true")
    parser.add_argument("--use-album-subfolders", action="store_true")
    parser.add_argument("--loop", type=int, help="Loop delay in minutes")

    return parser.parse_args()


def SpotiFLAC(url, output_dir, services=["tidal"], filename_format="{title} - {artist}", use_track_numbers=False, use_artist_subfolders=False, use_album_subfolders=False, loop=None):
    check_for_updates()
    global config
    config = Config(url, output_dir, services, filename_format, use_track_numbers, use_artist_subfolders, use_album_subfolders, False, False, False, "", [], None, loop)
    try:
        fetch_tracks(config.url)
        download_tracks(range(len(config.tracks)))
    except KeyboardInterrupt:
        print("\n\n[!] Download stopped by user. Partial files were cleaned up.")


def main():
    args = parse_args()
    SpotiFLAC(args.url, args.output_dir, args.service, args.filename_format, args.use_track_numbers, args.use_artist_subfolders, args.use_album_subfolders, args.loop)


if __name__ == "__main__":
    main()


import time
import threading
from enum import Enum
from dataclasses import dataclass

class DownloadStatus(Enum):
    QUEUED = "queued"
    DOWNLOADING = "downloading"
    COMPLETED = "completed"
    FAILED = "failed"
    SKIPPED = "skipped"

@dataclass
class DownloadItem:
    id: str
    track_name: str
    artist_name: str
    album_name: str
    spotify_id: str
    status: DownloadStatus = DownloadStatus.QUEUED
    progress: float = 0.0
    total_size: float = 0.0
    speed: float = 0.0
    start_time: float = 0.0
    end_time: float = 0.0
    error_message: str = ""
    file_path: str = ""

class DownloadManager:
    """
    Equivalente ao estado global com Mutex (sync.RWMutex) do Go.
    Usamos o padrão Singleton para manter o estado global em todo o módulo.
    """
    _instance = None
    _creation_lock = threading.Lock()

    def __new__(cls):
        with cls._creation_lock:
            if cls._instance is None:
                cls._instance = super(DownloadManager, cls).__new__(cls)
                cls._instance._init_state()
        return cls._instance

    def _init_state(self):
        self.lock = threading.RLock()
        self.queue: list[DownloadItem] = []
        self.is_downloading = False
        self.current_speed = 0.0
        self.total_downloaded = 0.0
        self.current_item_id = ""
        self.session_start_time = 0.0

    def add_to_queue(self, item_id, track_name, artist_name, album_name, spotify_id):
        with self.lock:
            item = DownloadItem(
                id=item_id,
                track_name=track_name,
                artist_name=artist_name,
                album_name=album_name,
                spotify_id=spotify_id
            )
            self.queue.append(item)
            if self.session_start_time == 0.0:
                self.session_start_time = time.time()

    def start_download(self, item_id):
        with self.lock:
            for item in self.queue:
                if item.id == item_id:
                    item.status = DownloadStatus.DOWNLOADING
                    item.start_time = time.time()
                    item.progress = 0.0
                    break
            self.current_item_id = item_id
            self.is_downloading = True

    def update_progress(self, item_id, progress_mb, speed_mbps):
        with self.lock:
            self.current_speed = speed_mbps
            for item in self.queue:
                if item.id == item_id:
                    item.progress = progress_mb
                    item.speed = speed_mbps
                    break

    def complete_download(self, item_id, filepath, final_size_mb):
        with self.lock:
            for item in self.queue:
                if item.id == item_id:
                    item.status = DownloadStatus.COMPLETED
                    item.end_time = time.time()
                    item.file_path = filepath
                    item.progress = final_size_mb
                    item.total_size = final_size_mb
                    
                    self.total_downloaded += final_size_mb
                    break
            self.is_downloading = False

    def fail_download(self, item_id, error_msg):
        with self.lock:
            for item in self.queue:
                if item.id == item_id:
                    item.status = DownloadStatus.FAILED
                    item.end_time = time.time()
                    item.error_message = error_msg
                    break
            self.is_downloading = False

    def get_queue_info(self) -> dict:
        with self.lock:
            queued = completed = failed = skipped = 0
            for item in self.queue:
                if item.status == DownloadStatus.QUEUED: queued += 1
                elif item.status == DownloadStatus.COMPLETED: completed += 1
                elif item.status == DownloadStatus.FAILED: failed += 1
                elif item.status == DownloadStatus.SKIPPED: skipped += 1

            return {
                "is_downloading": self.is_downloading,
                "current_speed": self.current_speed,
                "total_downloaded": self.total_downloaded,
                "queued_count": queued,
                "completed_count": completed,
                "failed_count": failed,
                "skipped_count": skipped,
                "queue": [vars(item) for item in self.queue]
            }

class RichProgressCallback:
    def __init__(self, item_id: str = ""):
        self.item_id = item_id
        self.start_time = time.time()
        self.last_time = self.start_time
        self.last_bytes = 0
        self.manager = DownloadManager()

    def __call__(self, current_bytes: int, total_bytes: int) -> None:
        now = time.time()
        time_diff = now - self.last_time
        
        # Atualiza a tela a cada 0.25s para não piscar (flickering)
        if time_diff >= 0.25 or current_bytes == total_bytes:
            bytes_diff = current_bytes - self.last_bytes
            
            # Calcula MB/s
            speed_mbps = 0.0
            if time_diff > 0:
                speed_mbps = (bytes_diff / (1024 * 1024)) / time_diff
                
            mb_downloaded = current_bytes / (1024 * 1024)

            # Print visual no terminal
            if total_bytes > 0:
                percent = (current_bytes / total_bytes) * 100
                print(f"\rDownloaded: {percent:.1f}% | {mb_downloaded:.2f} MB | {speed_mbps:.2f} MB/s", end="", flush=True)
            else:
                print(f"\rDownloaded: {mb_downloaded:.2f} MB | {speed_mbps:.2f} MB/s", end="", flush=True)

            # Atualiza o estado global
            if self.item_id:
                self.manager.update_progress(self.item_id, mb_downloaded, speed_mbps)

            # Reseta os contadores para o próximo ciclo
            self.last_time = now
            self.last_bytes = current_bytes

        if current_bytes == total_bytes and total_bytes > 0:
            print()


import requests
import asyncio
from mutagen.flac import FLAC
import os

class DeezerDownloader:
    def __init__(self):
        self.session = requests.Session()
        self.session.headers.update({
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/145.0.0.0 Safari/537.36"
        })
        self.progress_callback = None

    def set_progress_callback(self, callback):
        self.progress_callback = callback

    def get_track_by_isrc(self, isrc):
        try:
            url = f"https://api.deezer.com/2.0/track/isrc:{isrc}"
            response = self.session.get(url)
            response.raise_for_status()

            data = response.json()

            if 'error' in data:
                print(f"Error from Deezer API: {data['error']['message']}")
                return None

            return data
        except requests.exceptions.RequestException as e:
            print(f"Error fetching track data: {e}")
            return None

    def extract_metadata(self, track_data):
        metadata = {}

        metadata['title'] = track_data.get('title', '')
        metadata['title_short'] = track_data.get('title_short', '')
        metadata['duration'] = track_data.get('duration', 0)
        metadata['track_position'] = track_data.get('track_position', 1)
        metadata['disk_number'] = track_data.get('disk_number', 1)
        metadata['isrc'] = track_data.get('isrc', '')
        metadata['release_date'] = track_data.get('release_date', '')
        metadata['explicit_lyrics'] = track_data.get('explicit_lyrics', False)

        if 'artist' in track_data:
            metadata['artist'] = track_data['artist'].get('name', '')
            metadata['artist_id'] = track_data['artist'].get('id', '')

        if 'contributors' in track_data:
            artists = []
            for contributor in track_data['contributors']:
                if contributor.get('role') == 'Main':
                    artists.append(contributor.get('name', ''))
            metadata['artists'] = ', '.join(artists) if artists else metadata.get('artist', '')

        if 'album' in track_data:
            album = track_data['album']
            metadata['album'] = album.get('title', '')
            metadata['album_id'] = album.get('id', '')
            metadata['cover_url'] = album.get('cover_xl', album.get('cover_big', ''))
            metadata['cover_md5'] = album.get('md5_image', '')

        metadata['deezer_link'] = track_data.get('link', '')
        metadata['preview_url'] = track_data.get('preview', '')

        return metadata

    def download_cover_art(self, cover_url, filename):
        if not cover_url:
            return None

        try:
            response = self.session.get(cover_url)
            response.raise_for_status()

            cover_path = f"{filename}_cover.jpg"
            with open(cover_path, 'wb') as f:
                f.write(response.content)

            return cover_path
        except Exception as e:
            print(f"Error downloading cover art: {e}")
            return None

    def embed_metadata(self, file_path, metadata, cover_path=None):
        try:
            audio = FLAC(file_path)

            audio.clear()

            if metadata.get('title'):
                audio['TITLE'] = metadata['title']
            if metadata.get('artists'):
                audio['ARTIST'] = metadata['artists']
            elif metadata.get('artist'):
                audio['ARTIST'] = metadata['artist']
            if metadata.get('album'):
                audio['ALBUM'] = metadata['album']
            if metadata.get('release_date'):
                audio['DATE'] = metadata['release_date']
            if metadata.get('track_position'):
                audio['TRACKNUMBER'] = str(metadata['track_position'])
            if metadata.get('disk_number'):
                audio['DISCNUMBER'] = str(metadata['disk_number'])
            if metadata.get('isrc'):
                audio['ISRC'] = metadata['isrc']

            if cover_path and os.path.exists(cover_path):
                with open(cover_path, 'rb') as f:
                    cover_data = f.read()

                from mutagen.flac import Picture
                picture = Picture()
                picture.type = 3
                picture.mime = 'image/jpeg'
                picture.desc = 'Cover'
                picture.data = cover_data
                audio.add_picture(picture)

            audio.save()
            print(f"Metadata embedded successfully in {file_path}")

        except Exception as e:
            print(f"Error embedding metadata: {e}")

    async def download_by_isrc(self, isrc, output_dir="."):
        print(f"Fetching track info for ISRC: {isrc}")

        track_data = self.get_track_by_isrc(isrc)
        if not track_data:
            print("Failed to get track data from Deezer API")
            return False

        metadata = self.extract_metadata(track_data)
        print(f"Found track: {metadata.get('artists', 'Unknown')} - {metadata.get('title', 'Unknown')}")

        track_id = track_data.get('id')
        if not track_id:
            print("No track ID found in Deezer API response")
            return False

        print(f"Using track ID: {track_id}")

        api_url = f"https://api.deezmate.com/dl/{track_id}"
        print(f"Requesting download links from: {api_url}")

        try:
            response = self.session.get(api_url)
            response.raise_for_status()
            api_data = response.json()

            if not api_data.get('success'):
                print("API request failed")
                return False

            links = api_data.get('links', {})
            flac_url = links.get('flac')

            if not flac_url:
                print("No FLAC download link found in API response")
                return False

            print(f"Successfully obtained FLAC download URL")

        except Exception as e:
            print(f"Error getting download URL from API: {e}")
            return False

        print("Downloading FLAC file...")
        try:
            response = self.session.get(flac_url)
            response.raise_for_status()

            safe_title = "".join(c for c in metadata.get('title', 'Unknown') if c.isalnum() or c in (' ', '-', '_')).rstrip()
            safe_artist = "".join(c for c in metadata.get('artists', 'Unknown') if c.isalnum() or c in (' ', '-', '_')).rstrip()
            filename = f"{safe_artist} - {safe_title}.flac"
            file_path = os.path.join(output_dir, filename)

            with open(file_path, 'wb') as f:
                f.write(response.content)

            downloaded = len(response.content)
            print(f"File size: {downloaded} bytes ({downloaded / (1024*1024):.2f} MB)")

            if self.progress_callback:
                self.progress_callback(downloaded, downloaded)

            print(f"Downloaded: {file_path}")

            cover_path = None
            if metadata.get('cover_url'):
                print("Downloading cover art...")
                cover_path = self.download_cover_art(metadata['cover_url'],
                                                   os.path.join(output_dir, f"{safe_artist} - {safe_title}"))

            print("Embedding metadata...")
            self.embed_metadata(file_path, metadata, cover_path)

            if cover_path and os.path.exists(cover_path):
                os.remove(cover_path)

            print(f"Successfully downloaded and tagged: {filename}")
            return True

        except Exception as e:
            print(f"Error downloading file: {e}")
            return False

async def main():
    print("=== DeezerDL - Deezer Downloader ===")
    downloader = DeezerDownloader()

    isrc = "USAT22409172"
    output_dir = "."

    success = await downloader.download_by_isrc(isrc, output_dir)
    if success:
        print("Download completed successfully!")
    else:
        print("Download failed!")

if __name__ == "__main__":
    try:
        import sys
        if sys.platform == "win32":
            import os
            os.system("chcp 65001 > nul")
            try:
                sys.stdout.reconfigure(encoding='utf-8')
            except:
                pass
    except:
        pass

    asyncio.run(main())


import base64
import json
import os
import random
import re
import subprocess
import time
import xml.etree.ElementTree as ET
from typing import Callable, Dict, List, Optional, Tuple
from urllib.parse import quote
from random import randrange
import requests
from mutagen.flac import FLAC, Picture
from mutagen.id3 import PictureType

def sanitize_filename(value: str) -> str:
    return re.sub(r'[\\/*?:"<>|]', "", value).strip()

def get_first_artist(artist_str: str) -> str:
    if not artist_str:
        return "Unknown"
    return artist_str.split(",")[0].strip()

def safe_int(value) -> int:
    try:
        return int(value)
    except (ValueError, TypeError):
        return 0
   
def get_random_user_agent():
    return f"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_{randrange(11, 15)}_{randrange(4, 9)}) AppleWebKit/{randrange(530, 537)}.{randrange(30, 37)} (KHTML, like Gecko) Chrome/{randrange(80, 105)}.0.{randrange(3000, 4500)}.{randrange(60, 125)} Safari/{randrange(530, 537)}.{randrange(30, 36)}"

def build_tidal_filename(title, artist, album, album_artist, release_date, track_number, disc_number, format_string, include_track_number, position, use_album_track_number):
    number_to_use = track_number if use_album_track_number and track_number > 0 else position
    year = release_date[:4] if len(release_date) >= 4 else ""

    if "{" in format_string:
        filename = (format_string.replace("{title}", title)
                    .replace("{artist}", artist)
                    .replace("{album}", album)
                    .replace("{album_artist}", album_artist)
                    .replace("{year}", year)
                    .replace("{date}", sanitize_filename(release_date)))
        
        if disc_number > 0:
            filename = filename.replace("{disc}", str(disc_number))
        else:
            filename = filename.replace("{disc}", "")
            
        if number_to_use > 0:
            filename = filename.replace("{track}", f"{number_to_use:02d}")
        else:
            filename = re.sub(r"\{track\}[\.\s-]*", "", filename)
    else:
        if format_string == "artist-title":
            filename = f"{artist} - {title}"
        elif format_string == "title":
            filename = title
        else:
            filename = f"{title} - {artist}"
        if include_track_number and position > 0:
            filename = f"{number_to_use:02d}. {filename}"

    return sanitize_filename(filename) + ".flac"

def parse_manifest(manifest_b64: str) -> Tuple[str, str, List[str], str]:
    try:
        manifest_bytes = base64.b64decode(manifest_b64)
    except Exception as exc:
        raise Exception(f"failed to decode manifest: {exc}")

    manifest_str = manifest_bytes.decode(errors="ignore").strip()
    
    # Formato BTS (JSON)
    if manifest_str.startswith("{"):
        try:
            data = json.loads(manifest_str)
            urls = data.get("urls", [])
            mime_type = data.get("mimeType", "")
            if urls:
                print(f"Manifest: BTS format ({mime_type})")
                return urls[0], "", [], mime_type
            raise Exception("no URLs in BTS manifest")
        except Exception as exc:
            raise Exception(f"failed to parse BTS manifest: {exc}")

    # Formato DASH (XML)
    print("Manifest: DASH format")
    init_url = ""
    media_template = ""
    segment_count = 0

    try:
        mpd = ET.fromstring(manifest_str)
        ns = {"mpd": mpd.tag.split("}")[0].strip("{")} if "}" in mpd.tag else {}
        seg_template = mpd.find(".//mpd:SegmentTemplate", ns)
        if seg_template is None:
            seg_template = mpd.find(".//SegmentTemplate")
            
        if seg_template is not None:
            init_url = seg_template.get("initialization", "")
            media_template = seg_template.get("media", "")
            timeline = seg_template.find("mpd:SegmentTimeline", ns) or seg_template.find("SegmentTimeline")
            
            if timeline is not None:
                segments = []
                for seg in timeline.findall("mpd:S", ns) or timeline.findall("S"):
                    repeat = int(seg.get("r") or 0)
                    segments.append(repeat + 1)
                segment_count = sum(segments)

    except Exception:
        pass

    # Regex Fallback caso o XML falhe
    if not init_url or not media_template or segment_count == 0:
        print("Using regex fallback for DASH manifest...")
        init_match = re.search(r'initialization="([^"]+)"', manifest_str)
        media_match = re.search(r'media="([^"]+)"', manifest_str)
        
        if init_match: init_url = init_match.group(1)
        if media_match: media_template = media_match.group(1)
        
        matches = re.findall(r'<S\s+[^>]*>', manifest_str)
        for match in matches:
            r_match = re.search(r'r="(\d+)"', match)
            repeat = int(r_match.group(1)) if r_match else 0
            segment_count += repeat + 1

    if not init_url:
        raise Exception("no initialization URL found in manifest")
    
    if segment_count == 0:
        raise Exception("no segments found in manifest")

    init_url = init_url.replace("&", "&")
    media_template = media_template.replace("&", "&")

    media_urls = []
    for i in range(1, segment_count + 1):
        media_urls.append(media_template.replace("$Number$", str(i)))

    return "", init_url, media_urls, ""


class TidalDownloader:
    def __init__(self, timeout: float = 15.0):
        self.session = requests.Session()
        self.session.timeout = timeout
        self.session.headers.update({
            "User-Agent": get_random_user_agent()
        })
        self.progress_callback: Callable[[int, int], None] = None
        
        self.apis = [
            "https://hifi-one.spotisaver.net",
            "https://hifi-two.spotisaver.net",
            "https://eu-central.monochrome.tf",
            "https://us-west.monochrome.tf",
            "https://api.monochrome.tf",
            "https://monochrome-api.samidy.com",
            "https://tidal.kinoplus.online"
        ]

    def set_progress_callback(self, callback: Callable[[int, int], None]) -> None:
        self.progress_callback = callback

    def get_tidal_url_from_spotify(self, spotify_track_id: str) -> str:

        print("Getting Tidal URL via Songlink HTML...")

        url = f"https://song.link/s/{spotify_track_id}"

        headers = {
            "User-Agent": "Mozilla/5.0",
            "Accept-Language": "en-US,en;q=0.9"
        }

        try:
            resp = self.session.get(url, headers=headers, timeout=10)
            resp.raise_for_status()

            html = resp.text

            match = re.search(
                r'https://listen\.tidal\.com/(track|album)/([0-9]+)',
                html
            )

            if not match:
                raise Exception("Tidal link not found in HTML")

            tidal_url = match.group(0)

            print(f"Found Tidal URL: {tidal_url}")

            return tidal_url

        except Exception as e:
            raise Exception(f"Error resolving Tidal URL: {e}")

    def get_track_id_from_url(self, tidal_url: str) -> int:
        parts = tidal_url.split("/track/")
        if len(parts) < 2:
            raise Exception("invalid tidal URL format")
        track_part = parts[1].split("?")[0].strip()
        try:
            return int(track_part)
        except ValueError as exc:
            raise Exception(f"failed to parse track ID: {exc}")

    def get_download_url_rotated(self, track_id: int, quality: str) -> Tuple[str, str]:
        if not self.apis:
            raise Exception("no APIs available")

        apis_shuffled = self.apis.copy()
        random.seed(time.time())
        random.shuffle(apis_shuffled)

        print(f"Rotating through {len(apis_shuffled)} APIs...")
        
        last_error = None
        for api_url in apis_shuffled:
            print(f"Trying API: {api_url}")
            url = f"{api_url}/track/?id={track_id}&quality={quality}"
            
            try:
                resp = self.session.get(url, timeout=15.0)
                if resp.status_code != 200:
                    last_error = f"HTTP {resp.status_code}"
                    continue
                
                body = resp.json()
                
                # Check v2 response first
                if isinstance(body, dict) and body.get("data", {}).get("manifest"):
                    print(f"✓ Success with: {api_url}")
                    return api_url, "MANIFEST:" + body["data"]["manifest"]
                
                # Check v1 response
                if isinstance(body, list):
                    for item in body:
                        if item.get("OriginalTrackUrl"):
                            print(f"✓ Success with: {api_url}")
                            return api_url, item["OriginalTrackUrl"]

                last_error = "no download URL or manifest in response"
                
            except Exception as e:
                last_error = str(e)
                continue

        raise Exception(f"all APIs failed. Last error: {last_error}")

    def _stream_download(self, url: str, filepath: str) -> None:
        temp_path = filepath + ".part"
        with self.session.get(url, stream=True, timeout=120) as resp:
            resp.raise_for_status()
            total = int(resp.headers.get("Content-Length") or 0)
            downloaded = 0
            with open(temp_path, "wb") as f:
                for chunk in resp.iter_content(chunk_size=256 * 1024):
                    if chunk:
                        f.write(chunk)
                        downloaded += len(chunk)
                        if self.progress_callback:
                            self.progress_callback(downloaded, total)
        os.replace(temp_path, filepath)

    def download_from_manifest(self, manifest_b64: str, output_path: str) -> None:
        direct_url, init_url, media_urls, mime_type = parse_manifest(manifest_b64)
        
        if direct_url and ("flac" in mime_type.lower() or not mime_type):
            print("Downloading file...")
            self._stream_download(direct_url, output_path)
            print("\nDownload complete")
            return

        temp_path = output_path + ".m4a.tmp"

        if direct_url:
            print(f"Downloading non-FLAC file ({mime_type})...")
            self._stream_download(direct_url, temp_path)
            print("\nDownload complete")
        else:
            total_segments = len(media_urls)
            print(f"Downloading {total_segments + 1} segments...")
            
            with open(temp_path, "wb") as f:
                print("Downloading init segment... ", end="", flush=True)
                resp = self.session.get(init_url)
                resp.raise_for_status()
                f.write(resp.content)
                print("OK")

                total_bytes = 0
                for i, media_url in enumerate(media_urls):
                    resp = self.session.get(media_url)
                    resp.raise_for_status()
                    f.write(resp.content)
                    total_bytes += len(resp.content)
                    
                    mb_downloaded = total_bytes / (1024 * 1024)
                    print(f"\rDownloading: {mb_downloaded:.2f} MB ({i+1}/{total_segments} segments)", end="")
            
            print(f"\nDownloaded: {os.path.getsize(temp_path) / (1024 * 1024):.2f} MB (Complete)")

        print("Converting to FLAC...")
        cmd = ["ffmpeg", "-y", "-i", temp_path, "-vn", "-c:a", "flac", output_path]
        
        # Ocultar janela no Windows
        si = None
        if os.name == 'nt':
            si = subprocess.STARTUPINFO()
            si.dwFlags |= subprocess.STARTF_USESHOWWINDOW

        result = subprocess.run(cmd, capture_output=True, text=True, startupinfo=si)
        
        if result.returncode != 0:
            m4a_path = output_path.replace(".flac", ".m4a")
            os.replace(temp_path, m4a_path)
            raise Exception(f"ffmpeg conversion failed (M4A saved as {os.path.basename(m4a_path)}): {result.stderr}")
            
        try: os.remove(temp_path)
        except: pass
        print("Download complete")

    def download_file(self, url: str, filepath: str) -> None:
        if url.startswith("MANIFEST:"):
            self.download_from_manifest(url.replace("MANIFEST:", "", 1), filepath)
        else:
            self._stream_download(url, filepath)

    def embed_metadata(self, filepath, title, artist, album, album_artist, date, track_num, total_tracks, 
                       disc_num, total_discs, cover_url, copyright, publisher, url):
        print("Embedding metadata and cover art...")
        try:
            cover_data = None
            if cover_url:
                try: 
                    resp = self.session.get(cover_url, timeout=15)
                    if resp.status_code == 200:
                        cover_data = resp.content
                except Exception as e:
                    print(f"Warning: Could not download cover: {e}")

            t_num = safe_int(track_num)
            t_total = safe_int(total_tracks)
            d_num = safe_int(disc_num)
            d_total = safe_int(total_discs)
            
            if t_num == 0: t_num = 1
            if d_num == 0: d_num = 1

            audio = FLAC(filepath)
            audio.delete()
            
            audio["TITLE"] = title
            audio["ARTIST"] = artist
            audio["ALBUM"] = album
            audio["ALBUMARTIST"] = album_artist
            audio["DATE"] = date
            audio["TRACKNUMBER"] = str(t_num)
            audio["TRACKTOTAL"] = str(t_total)
            audio["DISCNUMBER"] = str(d_num)
            audio["DISCTOTAL"] = str(d_total)
            if copyright: audio["COPYRIGHT"] = copyright
            if publisher: audio["ORGANIZATION"] = publisher
            if url: audio["URL"] = url
            audio["DESCRIPTION"] = "https://github.com/ShuShuzinhuu/SpotiFLAC-Module-Version"

            if cover_data:
                pic = Picture()
                pic.data = cover_data
                pic.type = PictureType.COVER_FRONT
                pic.mime = "image/jpeg"
                audio.add_picture(pic)
            audio.save()
            print("Metadata embedded successfully")

        except Exception as e:
            print(f"Warning: Failed to embed metadata: {e}")

    def download_by_url(self, tidal_url: str, output_dir: str, quality: str, filename_format: str, 
                        include_track_number: bool, position: int, 
                        spotify_track_name: str, spotify_artist_name: str, spotify_album_name: str, 
                        spotify_album_artist: str, spotify_release_date: str, spotify_cover_url: str, 
                        spotify_track_number: int, spotify_disc_number: int, spotify_total_tracks: int, 
                        embed_max_quality_cover: bool, spotify_total_discs: int, spotify_copyright: str, 
                        spotify_publisher: str, spotify_url: str, use_album_track_number: bool = False,
                        use_first_artist_only: bool = False, allow_fallback: bool = True):
        
        os.makedirs(output_dir, exist_ok=True)
        print(f"Using Tidal URL: {tidal_url}")

        track_id = self.get_track_id_from_url(tidal_url)

        artist_to_use = get_first_artist(spotify_artist_name) if use_first_artist_only else spotify_artist_name
        album_artist_to_use = get_first_artist(spotify_album_artist) if use_first_artist_only else spotify_album_artist

        filename = build_tidal_filename(
            sanitize_filename(spotify_track_name), 
            sanitize_filename(artist_to_use), 
            sanitize_filename(spotify_album_name), 
            sanitize_filename(album_artist_to_use),
            spotify_release_date, spotify_track_number, spotify_disc_number, 
            filename_format, include_track_number, position, use_album_track_number
        )
        filepath = os.path.join(output_dir, filename)

        if os.path.exists(filepath) and os.path.getsize(filepath) > 0:
            size_mb = os.path.getsize(filepath) / (1024 * 1024)
            print(f"File already exists: {filepath} ({size_mb:.2f} MB)")
            return filepath

        try:
            success_api, download_url = self.get_download_url_rotated(track_id, quality)
        except Exception as e:
            if quality == "HI_RES" and allow_fallback:
                print("⚠ HI_RES unavailable/failed on all APIs, falling back to LOSSLESS...")
                try:
                    success_api, download_url = self.get_download_url_rotated(track_id, "LOSSLESS")
                except Exception as e2:
                    raise Exception(f"failed to get download URL (HI_RES & LOSSLESS both failed): {e2}")
            else:
                raise e

        print(f"Downloading to: {filepath}")
        self.download_file(download_url, filepath)

        self.embed_metadata(filepath, spotify_track_name, spotify_artist_name, spotify_album_name, 
                            spotify_album_artist, spotify_release_date, spotify_track_number, 
                            spotify_total_tracks, spotify_disc_number, spotify_total_discs, 
                            spotify_cover_url, spotify_copyright, spotify_publisher, spotify_url)

        print("Done\n✓ Downloaded successfully from Tidal")
        return filepath

    def download_by_spotify_id(self, spotify_track_id, **kwargs):
        tidal_url = self.get_tidal_url_from_spotify(spotify_track_id)
        spotify_url = f"https://open.spotify.com/track/{spotify_track_id}"
        
        default_kwargs = {
            "output_dir": ".", "quality": "LOSSLESS", "filename_format": "{title} - {artist}",
            "include_track_number": False, "position": 1,
            "spotify_track_name": "Unknown", "spotify_artist_name": "Unknown", 
            "spotify_album_name": "Unknown", "spotify_album_artist": "Unknown",
            "spotify_release_date": "", "spotify_cover_url": "", "spotify_track_number": 1,
            "spotify_disc_number": 1, "spotify_total_tracks": 1, "embed_max_quality_cover": True,
            "spotify_total_discs": 1, "spotify_copyright": "", "spotify_publisher": "", "spotify_url": spotify_url,
            "use_album_track_number": False, "use_first_artist_only": False, "allow_fallback": True
        }

        for key in kwargs:
            if key in default_kwargs:
                default_kwargs[key] = kwargs[key]

        return self.download_by_url(tidal_url, **default_kwargs)


import base64
import json
import os
import re
import subprocess
from typing import Callable, Dict
from urllib.parse import quote
import requests
from mutagen.flac import FLAC, Picture
from mutagen.id3 import PictureType
from mutagen.mp4 import MP4, MP4Cover

class ProgressCallback:
    def __call__(self, current: int, total: int) -> None:
        if total > 0:
            percent = (current / total) * 100
            print(f"\r{percent:.2f}% ({current}/{total})", end="")
        else:
            print(f"\r{current / (1024 * 1024):.2f} MB", end="")

def sanitize_filename(value: str) -> str:
    return re.sub(r'[\\/*?:"<>|]', "", value).strip()

def get_first_artist(artist_str: str) -> str:
    if not artist_str:
        return "Unknown"
    return artist_str.split(",")[0].strip()

def get_ffmpeg_path() -> str:
    return "ffmpeg"

def get_ffprobe_path() -> str:
    return "ffprobe"

def safe_int(value) -> int:
    try:
        return int(value)
    except (ValueError, TypeError):
        return 0

class AmazonDownloader:
    def __init__(self, timeout: float = 120.0):
        self.session = requests.Session()
        self.session.timeout = timeout
        self.session.headers.update({
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/145.0.0.0 Safari/537.36"
        })
        self.progress_callback: Callable[[int, int], None] = ProgressCallback()

    def set_progress_callback(self, callback: Callable[[int, int], None]) -> None:
        self.progress_callback = callback

    def get_amazon_url_from_spotify(self, spotify_track_id: str) -> str:
        print("Getting Amazon URL via Songlink HTML...")

        url = f"https://song.link/s/{spotify_track_id}"

        headers = {
            "User-Agent": "Mozilla/5.0",
            "Accept-Language": "en-US,en;q=0.9"
        }

        try:
            resp = self.session.get(url, headers=headers, timeout=10)
            resp.raise_for_status()

            html = resp.text

            match = re.search(
                r'https://music\.amazon\.com/(tracks|albums)/([A-Z0-9]{10})',
                html
            )

            if not match:
                raise Exception("Amazon link not found in HTML")

            asin = match.group(2)

            base = base64.b64decode(
                "aHR0cHM6Ly9tdXNpYy5hbWF6b24uY29tL3RyYWNrcy8="
            ).decode()

            amazon_url = f"{base}{asin}?musicTerritory=US"

            print(f"Found Amazon URL: {amazon_url}")

            return amazon_url

        except Exception as e:
            raise Exception(f"Error resolving Amazon URL: {e}")

    def _get_codec(self, filepath: str) -> str:
        try:
            cmd = [
                get_ffprobe_path(), "-v", "quiet", "-select_streams", "a:0",
                "-show_entries", "stream=codec_name", "-of", "default=noprint_wrappers=1:nokey=1",
                filepath
            ]
            si = None
            if os.name == 'nt':
                si = subprocess.STARTUPINFO()
                si.dwFlags |= subprocess.STARTF_USESHOWWINDOW

            codec = subprocess.check_output(cmd, text=True, startupinfo=si).strip()
            return codec
        except:
            return "m4a"

    def download_from_afkar_xyz(self, amazon_url: str, output_dir: str) -> str:
        asin_match = re.search(r'(B[0-9A-Z]{9})', amazon_url)
        if not asin_match:
            raise Exception(f"Failed to extract ASIN from: {amazon_url}")
        asin = asin_match.group(1)

        api_url = f"https://amzn.afkarxyz.fun/api/track/{asin}"
        print(f"Fetching from Amazon API (ASIN: {asin})...")
        
        resp = self.session.get(api_url)
        if resp.status_code != 200:
             raise Exception(f"Amazon API returned status {resp.status_code}")

        data = resp.json()
        stream_url = data.get("streamUrl")
        decryption_key = data.get("decryptionKey")

        if not stream_url:
            raise Exception("No stream URL found in API response")

        temp_file = os.path.join(output_dir, f"{asin}.enc")
        print(f"Downloading track...")
        
        with self.session.get(stream_url, stream=True) as r:
            r.raise_for_status()
            total = int(r.headers.get("Content-Length") or 0)
            downloaded = 0
            with open(temp_file, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
                        downloaded += len(chunk)
                        if self.progress_callback:
                            self.progress_callback(downloaded, total)
        print()

        if decryption_key:
            print("Decrypting file...")
            codec = self._get_codec(temp_file)
            ext = ".flac" if codec == "flac" else ".m4a"
            decrypted_path = os.path.join(output_dir, f"{asin}{ext}")

            cmd = [
                get_ffmpeg_path(), "-y",
                "-decryption_key", decryption_key.strip(),
                "-i", temp_file,
                "-c", "copy",
                decrypted_path
            ]
            
            si = None
            if os.name == 'nt':
                si = subprocess.STARTUPINFO()
                si.dwFlags |= subprocess.STARTF_USESHOWWINDOW

            result = subprocess.run(cmd, capture_output=True, startupinfo=si)
            if result.returncode != 0:
                os.remove(temp_file)
                raise Exception(f"Decryption failed: {result.stderr.decode()}")
            
            os.remove(temp_file)
            return decrypted_path
        
        final_path = os.path.join(output_dir, f"{asin}.m4a")
        if os.path.exists(final_path):
             os.remove(final_path)
        os.rename(temp_file, final_path)
        return final_path

    def download_by_url(self, amazon_url: str, output_dir: str, quality: str, filename_format: str, 
                        playlist_name: str, playlist_owner: str, include_track_number: bool, position: int, 
                        spotify_track_name: str, spotify_artist_name: str, spotify_album_name: str, 
                        spotify_album_artist: str, spotify_release_date: str, spotify_cover_url: str, 
                        spotify_track_number: int, spotify_disc_number: int, spotify_total_tracks: int, 
                        embed_max_quality_cover: bool, spotify_total_discs: int, spotify_copyright: str, 
                        spotify_publisher: str, spotify_url: str, use_album_track_number: bool = False,
                        use_first_artist_only: bool = False):
        
        os.makedirs(output_dir, exist_ok=True)
        print(f"Using Amazon URL: {amazon_url}")

        # 1. Prepara os nomes otimizados
        artist_to_use = get_first_artist(spotify_artist_name) if use_first_artist_only else spotify_artist_name
        album_artist_to_use = get_first_artist(spotify_album_artist) if use_first_artist_only else spotify_album_artist

        safe_title = sanitize_filename(spotify_track_name)
        safe_artist = sanitize_filename(artist_to_use)
        safe_album = sanitize_filename(spotify_album_name)
        safe_album_artist = sanitize_filename(album_artist_to_use)
        year = spotify_release_date[:4] if len(spotify_release_date) >= 4 else ""

        track_num_for_filename = position
        if use_album_track_number and safe_int(spotify_track_number) > 0:
            track_num_for_filename = safe_int(spotify_track_number)

        if "{" in filename_format:
            new_name = (filename_format.replace("{title}", safe_title)
                        .replace("{artist}", safe_artist)
                        .replace("{album}", safe_album)
                        .replace("{album_artist}", safe_album_artist)
                        .replace("{year}", year))
            
            new_name = new_name.replace("{disc}", str(spotify_disc_number) if safe_int(spotify_disc_number) > 0 else "")
            
            if track_num_for_filename > 0:
                new_name = new_name.replace("{track}", f"{track_num_for_filename:02d}")
            else:
                new_name = re.sub(r'\{track\}[\.\s-]*', "", new_name)
        else:
            if filename_format == "artist-title":
                new_name = f"{safe_artist} - {safe_title}"
            elif filename_format == "title":
                new_name = safe_title
            else:
                new_name = f"{safe_title} - {safe_artist}"
            
            if include_track_number and track_num_for_filename > 0:
                new_name = f"{track_num_for_filename:02d}. {new_name}"

        new_name = sanitize_filename(new_name)
        
        # Checa se o arquivo FLAC ou M4A já existe no disco antes de acionar a API
        expected_path_flac = os.path.join(output_dir, new_name + ".flac")
        expected_path_m4a = os.path.join(output_dir, new_name + ".m4a")

        if os.path.exists(expected_path_flac) and os.path.getsize(expected_path_flac) > 0:
            size_mb = os.path.getsize(expected_path_flac) / (1024 * 1024)
            print(f"File already exists: {expected_path_flac} ({size_mb:.2f} MB)")
            return expected_path_flac
        
        if os.path.exists(expected_path_m4a) and os.path.getsize(expected_path_m4a) > 0:
            size_mb = os.path.getsize(expected_path_m4a) / (1024 * 1024)
            print(f"File already exists: {expected_path_m4a} ({size_mb:.2f} MB)")
            return expected_path_m4a

        # 3. Baixa o arquivo apenas se não existir
        file_path = self.download_from_afkar_xyz(amazon_url, output_dir)
        ext = os.path.splitext(file_path)[1] or ".flac"
        new_path = os.path.join(output_dir, new_name + ext)
        
        if os.path.exists(new_path):
             try: os.remove(new_path)
             except: pass
             
        os.replace(file_path, new_path)
        print(f"Renamed to: {new_name + ext}")

        self.embed_metadata(new_path, spotify_track_name, spotify_artist_name, spotify_album_name, 
                            spotify_album_artist, spotify_release_date, spotify_track_number, 
                            spotify_total_tracks, spotify_disc_number, spotify_total_discs, 
                            spotify_cover_url, spotify_copyright, spotify_publisher, spotify_url)

        print("Done\n✓ Downloaded successfully from Amazon Music")
        return new_path

    def embed_metadata(self, filepath, title, artist, album, album_artist, date, track_num, total_tracks, 
                       disc_num, total_discs, cover_url, copyright, publisher, url):
        print("Embedding metadata and cover art...")
        try:
            cover_data = None
            if cover_url:
                try: 
                    resp = self.session.get(cover_url, timeout=15)
                    if resp.status_code == 200:
                        cover_data = resp.content
                except Exception as e:
                    print(f"Warning: Could not download cover: {e}")

            t_num = safe_int(track_num)
            t_total = safe_int(total_tracks)
            d_num = safe_int(disc_num)
            d_total = safe_int(total_discs)
            
            if t_num == 0: t_num = 1
            if d_num == 0: d_num = 1

            if filepath.endswith(".flac"):
                audio = FLAC(filepath)
                audio.delete()
                
                audio["TITLE"] = title
                audio["ARTIST"] = artist
                audio["ALBUM"] = album
                audio["ALBUMARTIST"] = album_artist
                audio["DATE"] = date
                audio["TRACKNUMBER"] = str(t_num)
                audio["TRACKTOTAL"] = str(t_total)
                audio["DISCNUMBER"] = str(d_num)
                audio["DISCTOTAL"] = str(d_total)
                if copyright: audio["COPYRIGHT"] = copyright
                if publisher: audio["ORGANIZATION"] = publisher
                if url: audio["URL"] = url
                audio["DESCRIPTION"] = "https://github.com/ShuShuzinhuu/SpotiFLAC-Module-Version"

                if cover_data:
                    pic = Picture()
                    pic.data = cover_data
                    pic.type = PictureType.COVER_FRONT
                    pic.mime = "image/jpeg"
                    audio.add_picture(pic)
                audio.save()

            elif filepath.endswith(".m4a"):
                try:
                    audio = MP4(filepath)
                except:
                    audio = MP4(filepath)
                
                audio.delete()
                
                audio["\xa9nam"] = title
                audio["\xa9ART"] = artist
                audio["\xa9alb"] = album
                audio["aART"] = album_artist
                audio["\xa9day"] = date
                audio["trkn"] = [(t_num, t_total)]
                audio["disk"] = [(d_num, d_total)]
                if copyright: audio["cprt"] = copyright
                
                if cover_data:
                    audio["covr"] = [MP4Cover(cover_data, imageformat=MP4Cover.FORMAT_JPEG)]
                audio.save()
                
            print("Metadata embedded successfully")

        except Exception as e:
            print(f"Warning: Failed to embed metadata: {e}")

    def download_by_spotify_id(self, spotify_track_id, **kwargs):
        amazon_url = self.get_amazon_url_from_spotify(spotify_track_id)
        
        default_kwargs = {
            "output_dir": ".", "quality": "LOSSLESS", "filename_format": "{title} - {artist}",
            "playlist_name": "", "playlist_owner": "", "include_track_number": False, "position": 1,
            "spotify_track_name": "Unknown", "spotify_artist_name": "Unknown", 
            "spotify_album_name": "Unknown", "spotify_album_artist": "Unknown",
            "spotify_release_date": "", "spotify_cover_url": "", "spotify_track_number": 1,
            "spotify_disc_number": 1, "spotify_total_tracks": 1, "embed_max_quality_cover": True,
            "spotify_total_discs": 1, "spotify_copyright": "", "spotify_publisher": "", "spotify_url": "",
            "use_album_track_number": False, "use_first_artist_only": False
        }

        for key in kwargs:
            if key in default_kwargs:
                default_kwargs[key] = kwargs[key]

        return self.download_by_url(amazon_url, **default_kwargs)


import importlib.metadata
import requests

def check_for_updates():
    package_name = "spotiflac" 
    
    try:
        current_version = importlib.metadata.version(package_name)
        
        resp = requests.get(f"https://pypi.org/pypi/{package_name}/json", timeout=2)
        
        if resp.status_code == 200:
            latest_version = resp.json()["info"]["version"]
            
            if current_version != latest_version:
                width = 68
                
                print(f"\n ╭" + "─" * (width-2) + "╮")
                
                title_line = f"  NEW VERSION AVAILABLE! ({current_version} -> {latest_version})"
                print(f" │{title_line.ljust(width-2)}│")
                
                print(f" ├" + "─" * (width-2) + "┤")
                
                mod_line = f"  Module: pip install -U {package_name}"
                app_line = f"  App:    https://github.com/ShuShuzinhuu/SpotiFLAC-Module-Version"
                
                print(f" │{mod_line.ljust(width-2)}│")
                print(f" │{app_line.ljust(width-2)}│")
                
                print(f" ╰" + "─" * (width-2) + "╯\n")
                
                
    except importlib.metadata.PackageNotFoundError:
        pass
    except Exception:
        pass


import os
import re
import requests
import time
from typing import Callable
from mutagen.flac import FLAC, Picture
from mutagen.id3 import PictureType

def sanitize_filename(value: str) -> str:
    return re.sub(r'[\\/*?:"<>|]', "", value).strip()

def safe_int(value) -> int:
    try:
        return int(value)
    except (ValueError, TypeError):
        return 0

class SpotiDownloader:

    _cached_token = None

    def __init__(self, timeout: float = 15.0):
        self.session = requests.Session()
        self.session.timeout = timeout
        self.session.headers.update({
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
        })
        self.progress_callback: Callable[[int, int], None] = None

    def set_progress_callback(self, callback: Callable[[int, int], None]) -> None:
        self.progress_callback = callback

    def fetch_token(self) -> str:
        if SpotiDownloader._cached_token:
            return SpotiDownloader._cached_token

        print("Fetching session token from SpotiDownloader...")
        url = "https://spdl.afkarxyz.fun/token"
        
        for attempt in range(1, 4):
            try:
                resp = self.session.get(url, timeout=10)
                resp.raise_for_status()
                data = resp.json()
                token = data.get("token")
                if token:
                    SpotiDownloader._cached_token = token
                    return token
            except Exception as e:
                if attempt == 3:
                    raise Exception(f"Failed to fetch token after 3 attempts: {e}")
                time.sleep(1)
                
        raise Exception("Token not found in response")

    def get_flac_download_link(self, track_id: str, token: str) -> str:
        print(f"Requesting FLAC download link for ID: {track_id}...")
        url = "https://api.spotidownloader.com/download"
        
        headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
            "Origin": "https://spotidownloader.com",
            "Referer": "https://spotidownloader.com/",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/155.0.0.0 Safari/537.36"
        }
        
        payload = {"id": track_id, "flac": True}
        
        resp = self.session.post(url, json=payload, headers=headers, timeout=15)
        resp.raise_for_status()
        data = resp.json()
        
        if not data.get("success"):
            raise Exception("Download API returned success=false")
            
        flac_link = data.get("linkFlac")
        standard_link = data.get("link")
        
        final_link = None

        if flac_link and ".flac" in flac_link:
            final_link = flac_link
        elif standard_link and ".flac" in standard_link:
            final_link = standard_link
            
        if not final_link:
            raise Exception("SpotiDownloader API did not return a FLAC link for this track (only standard quality available).")
            
        return final_link
            


    def _stream_download(self, url: str, filepath: str, token: str) -> None:
        temp_path = filepath + ".part"
        
        # O download final desta API exige os headers de auth!
        headers = {
            "Authorization": f"Bearer {token}",
            "Origin": "https://spotidownloader.com",
            "Referer": "https://spotidownloader.com/"
        }
        
        with self.session.get(url, headers=headers, stream=True, timeout=120) as resp:
            resp.raise_for_status()
            total = int(resp.headers.get("Content-Length") or 0)
            downloaded = 0
            with open(temp_path, "wb") as f:
                for chunk in resp.iter_content(chunk_size=256 * 1024):
                    if chunk:
                        f.write(chunk)
                        downloaded += len(chunk)
                        if self.progress_callback:
                            self.progress_callback(downloaded, total)
                            
        os.replace(temp_path, filepath)

    def download_by_spotify_id(self, spotify_track_id, **kwargs):
        output_dir = kwargs.get("output_dir", ".")
        os.makedirs(output_dir, exist_ok=True)
        
        # Limpa o ID caso venha sujo
        track_id = spotify_track_id.split("/")[-1].split("?")[0]
        
        # Formata o nome do arquivo
        safe_title = sanitize_filename(kwargs.get("spotify_track_name", "Unknown"))
        safe_artist = sanitize_filename(kwargs.get("spotify_artist_name", "Unknown").split(",")[0])
        
        # Extensão forçada para FLAC
        expected_filename = f"{safe_artist} - {safe_title}.flac"
        expected_path = os.path.join(output_dir, expected_filename)

        if os.path.exists(expected_path) and os.path.getsize(expected_path) > 0:
            print(f"File already exists: {expected_path}")
            return expected_path

        # Fluxo de requisições
        token = self.fetch_token()
        flac_url = self.get_flac_download_link(track_id, token)

        print("Downloading FLAC file from SpotiDownloader...")
        self._stream_download(flac_url, expected_path, token)
        print() # Quebra de linha após o progresso
        
        # Injeta as tags
        self.embed_metadata(
            expected_path, 
            kwargs.get("spotify_track_name"), kwargs.get("spotify_artist_name"),
            kwargs.get("spotify_album_name"), kwargs.get("spotify_album_artist"),
            kwargs.get("spotify_release_date"), kwargs.get("spotify_track_number", 1),
            kwargs.get("spotify_total_tracks", 1), kwargs.get("spotify_disc_number", 1),
            kwargs.get("spotify_total_discs", 1), kwargs.get("spotify_cover_url"),
            kwargs.get("spotify_copyright"), kwargs.get("spotify_publisher"), kwargs.get("spotify_url")
        )

        return expected_path

    def embed_metadata(self, filepath, title, artist, album, album_artist, date, track_num, total_tracks, 
                       disc_num, total_discs, cover_url, copyright, publisher, url):
        print("Embedding metadata and cover art...")
        try:
            cover_data = None
            if cover_url:
                try: 
                    resp = self.session.get(cover_url, timeout=15)
                    if resp.status_code == 200: cover_data = resp.content
                except Exception as e:
                    print(f"Warning: Could not download cover: {e}")

            t_num = safe_int(track_num) or 1
            t_total = safe_int(total_tracks) or 1
            d_num = safe_int(disc_num) or 1
            d_total = safe_int(total_discs) or 1

            audio = FLAC(filepath)
            audio.delete()
            
            audio["TITLE"] = title
            audio["ARTIST"] = artist
            audio["ALBUM"] = album
            audio["ALBUMARTIST"] = album_artist
            audio["DATE"] = date
            audio["TRACKNUMBER"] = str(t_num)
            audio["TRACKTOTAL"] = str(t_total)
            audio["DISCNUMBER"] = str(d_num)
            audio["DISCTOTAL"] = str(d_total)
            if copyright: audio["COPYRIGHT"] = copyright
            if publisher: audio["ORGANIZATION"] = publisher
            if url: audio["URL"] = url
            audio["DESCRIPTION"] = "https://github.com/ShuShuzinhuu/SpotiFLAC-Module-Version"

            if cover_data:
                pic = Picture()
                pic.data = cover_data
                pic.type = PictureType.COVER_FRONT
                pic.mime = "image/jpeg"
                audio.add_picture(pic)
            
            audio.save()
            print("Metadata embedded successfully")

        except Exception as e:
            print(f"Warning: Failed to embed metadata: {e}")


"""
Launcher script for SpotiFLAC CLI
"""
import argparse
import sys
import os
from SpotiFLAC.SpotiFLAC import SpotiFLAC

if getattr(sys, 'frozen', False):
    application_path = sys._MEIPASS
else:
    application_path = os.path.dirname(os.path.abspath(__file__))

sys.path.insert(0, application_path)

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("url", help="Spotify URL")
    parser.add_argument("output_dir", help="Output directory")
    parser.add_argument(
        "--service",
        choices=["tidal", "qobuz", "amazon","spoti", "youtube"],
        nargs="+",
        default=["tidal", "spoti", "qobuz", "amazon"],
        help="One or more services to try in order",
    )
    parser.add_argument(
        "--filename-format",
        default="{title} - {artist}",
        help="Custom filename format using placeholders (see examples below)"
    )
    parser.add_argument("--use-track-numbers", action="store_true", help="(Deprecated - use {track} in format)")
    parser.add_argument("--use-artist-subfolders", action="store_true")
    parser.add_argument("--use-album-subfolders", action="store_true")
    parser.add_argument("--loop", type=int, help="Loop delay in minutes")
    return parser.parse_args()

def main():
    args = parse_args()
    SpotiFLAC(
        args.url, 
        args.output_dir, 
        args.service, 
        args.filename_format, 
        args.use_track_numbers, 
        args.use_artist_subfolders, 
        args.use_album_subfolders, 
        args.loop
    )

if __name__ == '__main__':
    main()


from time import sleep
from urllib.parse import urlparse, parse_qs
import requests
import json
import base64
from random import randrange
from typing import Dict, Any, List, Tuple


def get_random_user_agent():
    return f"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_{randrange(11, 15)}_{randrange(4, 9)}) AppleWebKit/{randrange(530, 537)}.{randrange(30, 37)} (KHTML, like Gecko) Chrome/{randrange(80, 155)}.0.{randrange(3000, 4500)}.{randrange(60, 125)} Safari/{randrange(530, 537)}.{randrange(30, 36)}"

CLIENT_ID = base64.b64decode("ODNlNDQzMGI0NzAwNDM0YmFhMjEyMjhhOWM3ZDExYzU=").decode('utf-8')
CLIENT_SECRET = base64.b64decode("OWJiOWUxMzFmZjI4NDI0Y2I2YTQyMGFmZGY0MWQ0NGE=").decode('utf-8')

token_url = 'https://accounts.spotify.com/api/token'
playlist_base_url = 'https://api.spotify.com/v1/playlists/{}'
album_base_url = 'https://api.spotify.com/v1/albums/{}'
track_base_url = 'https://api.spotify.com/v1/tracks/{}'
artist_base_url = 'https://api.spotify.com/v1/artists/{}'
artist_albums_url = 'https://api.spotify.com/v1/artists/{}/albums'

headers = {
    'User-Agent': get_random_user_agent(),
    'Accept': 'application/json',
    'Accept-Language': 'en-US,en;q=0.9',
    'Accept-Encoding': 'gzip, deflate, br',
}


class SpotifyInvalidUrlException(Exception):
    pass


class SpotifyWebsiteParserException(Exception):
    pass


def parse_uri(uri):
    u = urlparse(uri)
    if u.netloc == "embed.spotify.com":
        if not u.query:
            raise SpotifyInvalidUrlException("ERROR: url {} is not supported".format(uri))
        qs = parse_qs(u.query)
        return parse_uri(qs['uri'][0])

    if not u.scheme and not u.netloc:
        return {"type": "playlist", "id": u.path}

    if u.scheme == "spotify":
        parts = uri.split(":")
    else:
        if u.netloc != "open.spotify.com" and u.netloc != "play.spotify.com":
            raise SpotifyInvalidUrlException("ERROR: url {} is not supported".format(uri))
        parts = u.path.split("/")
        if parts[1] == "embed":
            parts = parts[1:]
        if len(parts) > 1 and parts[1].startswith("intl-"):
            parts = parts[1:]

    l = len(parts)
    if l == 3 and parts[1] in ["album", "track", "playlist", "artist"]:
        return {"type": parts[1], "id": parts[2]}
    if l == 5 and parts[3] == "playlist":
        return {"type": parts[3], "id": parts[4]}
    if l >= 4 and parts[1] == "artist" and len(parts) >= 4:
        if parts[3] == "discography":
            discography_type = "all"
            if len(parts) >= 5 and parts[4] in ["all", "album", "single", "compilation"]:
                discography_type = parts[4]
            return {"type": "artist_discography", "id": parts[2], "discography_type": discography_type}
        else:
            return {"type": "artist", "id": parts[2]}

    raise SpotifyInvalidUrlException("ERROR: unable to determine Spotify URL type or type is unsupported.")


def get_json_from_api(api_url, access_token):
    request_headers = headers.copy()
    request_headers['Authorization'] = f'Bearer {access_token}'

    req = requests.get(api_url, headers=request_headers, timeout=10)

    if req.status_code == 429:
        seconds = int(req.headers.get("Retry-After", "5")) + 1
        print(f"INFO: rate limited! Sleeping for {seconds} seconds")
        sleep(seconds)
        return None

    if req.status_code != 200:
        raise SpotifyWebsiteParserException(f"ERROR: {api_url} gave us not a 200. Instead: {req.status_code}")

    return req.json()


def get_access_token():
    try:
        auth_str = f"{CLIENT_ID}:{CLIENT_SECRET}"
        auth_bytes = auth_str.encode('utf-8')
        auth_b64 = base64.b64encode(auth_bytes).decode('utf-8')

        token_headers = {
            'Authorization': f'Basic {auth_b64}',
            'Content-Type': 'application/x-www-form-urlencoded',
            'User-Agent': get_random_user_agent()
        }

        token_data = {
            'grant_type': 'client_credentials'
        }

        req = requests.post(token_url, headers=token_headers, data=token_data, timeout=10)

        if req.status_code != 200:
            print(f"Token request failed: {req.status_code}")
            print(f"Response: {req.text}")
            return {"error": f"Failed to get access token. Status code: {req.status_code}"}

        response = req.json()
        return {"accessToken": response.get("access_token")}

    except Exception as e:
        return {"error": f"Failed to get access token: {str(e)}"}


def fetch_tracks_in_batches(url: str, access_token: str, batch_size: int = 100, delay: float = 1.0) -> Tuple[
    List[Dict[str, Any]], int]:
    all_tracks = []
    current_batch = 0

    while url:
        print(f"Batch : {current_batch}")
        url_parts = url.split("offset=")
        if len(url_parts) > 1:
            offset_part = url_parts[1].split("&")[0]
            print(f"Offset : {offset_part}")
        print("-------------")

        track_data = get_json_from_api(url, access_token)
        if not track_data:
            break

        items = track_data.get('items', [])
        all_tracks.extend(items)
        url = track_data.get('next')

        if url and "&locale=" in url:
            url = url.split("&locale=")[0]

        if url and delay > 0:
            sleep(delay)

        current_batch += 1

    return all_tracks, current_batch


def get_raw_spotify_data(spotify_url, batch: bool = False, delay: float = 1.0):
    url_info = parse_uri(spotify_url)
    token = get_access_token()

    if "error" in token:
        return token

    access_token = token["accessToken"]
    raw_data = {}

    if url_info['type'] == "playlist":
        try:
            playlist_data = get_json_from_api(
                playlist_base_url.format(url_info["id"]),
                access_token
            )
            if not playlist_data:
                return {"error": "Failed to get playlist data"}

            raw_data = playlist_data
            total_tracks = playlist_data.get('tracks', {}).get('total', 0)

            if batch:
                tracks_url = f'https://api.spotify.com/v1/playlists/{url_info["id"]}/tracks?limit=100'
                tracks, num_batches = fetch_tracks_in_batches(tracks_url, access_token, 100, delay)
                raw_data['tracks']['items'] = tracks
                raw_data['_batch_count'] = num_batches
                raw_data['_batch_enabled'] = True

                if len(tracks) < total_tracks:
                    last_offset = len(tracks)
                    remaining_tracks = []
                    while last_offset < total_tracks:
                        print(f"Batch : {num_batches}")
                        print(f"Offset : {last_offset}")
                        print("-------------")
                        remainder_url = f'https://api.spotify.com/v1/playlists/{url_info["id"]}/tracks?offset={last_offset}&limit=100'
                        track_data = get_json_from_api(remainder_url, access_token)
                        if not track_data or not track_data.get('items'):
                            break
                        items = track_data.get('items', [])
                        remaining_tracks.extend(items)
                        if len(items) < 100:
                            break
                        last_offset += len(items)
                        num_batches += 1
                        if delay > 0:
                            sleep(delay)
                    tracks.extend(remaining_tracks)
                    raw_data['tracks']['items'] = tracks
                    raw_data['_batch_count'] = num_batches
            else:
                tracks = []
                tracks_url = f'https://api.spotify.com/v1/playlists/{url_info["id"]}/tracks?limit=100'
                while tracks_url:
                    track_data = get_json_from_api(tracks_url, access_token)
                    if not track_data:
                        break
                    tracks.extend(track_data['items'])
                    tracks_url = track_data.get('next')
                    if tracks_url and "&locale=" in tracks_url:
                        tracks_url = tracks_url.split("&locale=")[0]
                raw_data['tracks']['items'] = tracks
                raw_data['_batch_enabled'] = False
        except Exception as e:
            return {"error": f"Failed to get playlist data: {str(e)}"}

    elif url_info["type"] == "album":
        try:
            album_data = get_json_from_api(
                album_base_url.format(url_info["id"]),
                access_token
            )
            if not album_data:
                return {"error": "Failed to get album data"}

            album_data['_token'] = access_token
            raw_data = album_data
            total_tracks = album_data.get('total_tracks', 0)

            if batch:
                tracks_url = f'{album_base_url.format(url_info["id"])}/tracks?limit=50'
                tracks, num_batches = fetch_tracks_in_batches(tracks_url, access_token, 50, delay)
                raw_data['tracks']['items'] = tracks
                raw_data['_batch_count'] = num_batches
                raw_data['_batch_enabled'] = True

                if len(tracks) < total_tracks:
                    last_offset = len(tracks)
                    remaining_tracks = []
                    while last_offset < total_tracks:
                        print(f"Batch : {num_batches}")
                        print(f"Offset : {last_offset}")
                        print("-------------")
                        remainder_url = f'{album_base_url.format(url_info["id"])}/tracks?offset={last_offset}&limit=50'
                        track_data = get_json_from_api(remainder_url, access_token)
                        if not track_data or not track_data.get('items'):
                            break
                        items = track_data.get('items', [])
                        remaining_tracks.extend(items)
                        if len(items) < 50:
                            break
                        last_offset += len(items)
                        num_batches += 1
                        if delay > 0:
                            sleep(delay)
                    tracks.extend(remaining_tracks)
                    raw_data['tracks']['items'] = tracks
                    raw_data['_batch_count'] = num_batches
            else:
                tracks = []
                tracks_url = f'{album_base_url.format(url_info["id"])}/tracks?limit=50'
                while tracks_url:
                    track_data = get_json_from_api(tracks_url, access_token)
                    if not track_data:
                        break
                    tracks.extend(track_data['items'])
                    tracks_url = track_data.get('next')
                    if tracks_url and "&locale=" in tracks_url:
                        tracks_url = tracks_url.split("&locale=")[0]
                raw_data['tracks']['items'] = tracks
                raw_data['_batch_enabled'] = False
        except Exception as e:
            return {"error": f"Failed to get album data: {str(e)}"}

    elif url_info["type"] == "track":
        try:
            track_data = get_json_from_api(
                track_base_url.format(url_info["id"]),
                access_token
            )
            if not track_data:
                return {"error": "Failed to get track data"}
            raw_data = track_data
        except Exception as e:
            return {"error": f"Failed to get track data: {str(e)}"}

    return raw_data


def format_track_data(track_data):
    artists = []
    for artist in track_data.get('artists', []):
        if artist.get('name') is None:
            artists.append('Unknown Artist')
        else:
            artists.append(artist['name'])

    image_url = track_data.get('album', {}).get('images', [{}])[0].get('url', '') if track_data.get('album', {}).get(
        'images') else ''
    isrc = track_data.get('external_ids', {}).get('isrc', '')

    return {
        "track": {
            "artists": ", ".join(artists),
            "name": track_data.get('name', ''),
            "album_name": track_data.get('album', {}).get('name', ''),
            "duration_ms": track_data.get('duration_ms', 0),
            "images": image_url,
            "release_date": track_data.get('album', {}).get('release_date', ''),
            "track_number": track_data.get('track_number', 0),
            "external_urls": track_data.get('external_urls', {}).get('spotify', ''),
            "isrc": isrc
        }
    }


def format_album_data(album_data):
    artists = []
    for artist in album_data.get('artists', []):
        if artist.get('name') is None:
            artists.append('Unknown Artist')
        else:
            artists.append(artist['name'])

    image_url = album_data.get('images', [{}])[0].get('url', '') if album_data.get('images') else ''
    track_list = []

    for track in album_data.get('tracks', {}).get('items', []):
        track_artists = []
        for artist in track.get('artists', []):
            if artist.get('name') is None:
                track_artists.append('Unknown Artist')
            else:
                track_artists.append(artist['name'])

        track_id = track.get('id', '')
        track_isrc = ''
        if track_id and album_data.get('_token'):
            try:
                full_track_data = get_json_from_api(
                    track_base_url.format(track_id),
                    album_data.get('_token')
                )
                if full_track_data:
                    track_isrc = full_track_data.get('external_ids', {}).get('isrc', '')
            except:
                pass

        track_list.append({
            "artists": ", ".join(track_artists),
            "name": track.get('name', ''),
            "album_name": album_data.get('name', ''),
            "duration_ms": track.get('duration_ms', 0),
            "images": image_url,
            "release_date": album_data.get('release_date', ''),
            "track_number": track.get('track_number', 0),
            "external_urls": track.get('external_urls', {}).get('spotify', ''),
            "isrc": track_isrc
        })

    album_info = {
        "total_tracks": album_data.get('total_tracks', 0),
        "name": album_data.get('name', ''),
        "release_date": album_data.get('release_date', ''),
        "artists": ", ".join(artists),
        "images": image_url
    }

    if album_data.get('_batch_enabled', False):
        album_info["batch"] = f"{album_data.get('_batch_count', 1)}"

    return {
        "album_info": album_info,
        "track_list": track_list
    }


def format_playlist_data(playlist_data):
    image_url = playlist_data.get('images', [{}])[0].get('url', '') if playlist_data.get('images') else ''
    track_list = []

    for item in playlist_data.get('tracks', {}).get('items', []):
        track = item.get('track', {})
        if not track:
            continue

        artists = []
        for artist in track.get('artists', []):
            if artist.get('name') is None:
                artists.append('Unknown Artist')
            else:
                artists.append(artist['name'])

        track_image = ''
        if track.get('album', {}).get('images'):
            track_image = track.get('album', {}).get('images', [{}])[0].get('url', '')

        track_isrc = track.get('external_ids', {}).get('isrc', '')

        track_list.append({
            "artists": ", ".join(artists),
            "name": track.get('name', ''),
            "album_name": track.get('album', {}).get('name', ''),
            "duration_ms": track.get('duration_ms', 0),
            "images": track_image,
            "release_date": track.get('album', {}).get('release_date', ''),
            "track_number": track.get('track_number', 0),
            "external_urls": track.get('external_urls', {}).get('spotify', ''),
            "isrc": track_isrc
        })

    playlist_info = {
        "name": playlist_data.get('name', 'Unknown Playlist'),
        "tracks": {"total": playlist_data.get('tracks', {}).get('total', 0)},
        "followers": {"total": playlist_data.get('followers', {}).get('total', 0)},
        "owner": {
            "display_name": playlist_data.get('owner', {}).get('display_name', 'Unknown Owner'),
            "images": image_url
        }
    }

    if playlist_data.get('_batch_enabled', False):
        playlist_info["batch"] = f"{playlist_data.get('_batch_count', 1)}"

    return {
        "playlist_info": playlist_info,
        "track_list": track_list
    }


def process_spotify_data(raw_data, data_type):
    if not raw_data or "error" in raw_data:
        return {"error": "Invalid data provided"}

    try:
        if data_type == "track":
            return format_track_data(raw_data)
        elif data_type == "album":
            return format_album_data(raw_data)
        elif data_type == "playlist":
            return format_playlist_data(raw_data)
        else:
            print("Invalid data type")
            return {"error": "Invalid data type"}
    except Exception as e:
        print("Error processing data")
        return {"error": f"Error processing data: {str(e)}"}


def get_filtered_data(spotify_url, batch=False, delay=1.0):
    raw_data = get_raw_spotify_data(spotify_url, batch=batch, delay=delay)
    if raw_data and "error" not in raw_data:
        url_info = parse_uri(spotify_url)
        filtered_data = process_spotify_data(raw_data, url_info['type'])
        return filtered_data
    return {"error": "Failed to get raw data"}


if __name__ == '__main__':
    playlist = "https://open.spotify.com/playlist/37i9dQZEVXbNG2KDcFcKOF"
    album = "https://open.spotify.com/album/6J84szYCnMfzEcvIcfWMFL"
    song = "https://open.spotify.com/track/7so0lgd0zP2Sbgs2d7a1SZ"

    print("Fetching playlist...")
    filtered_playlist = get_filtered_data(playlist, batch=True, delay=0.1)
    print(json.dumps(filtered_playlist, indent=2))

    print("\nFetching album...")
    filtered_album = get_filtered_data(album)
    print(json.dumps(filtered_album, indent=2))

    print("\nFetching track...")
    filtered_track = get_filtered_data(song)
    print(json.dumps(filtered_track, indent=2))